import random

from mofun.helpers import remove_duplicates, position_index_farthest_from_axis, guess_elements_from_masses
from tests.fixtures import *

def test_remove_duplicates__leaves_order_untouched():
    assert remove_duplicates([(3, 2, 1)]) == [(3, 2, 1)]
    assert remove_duplicates([(3, 2, 1), (1, 2, 3)]) == [(3, 2, 1)]

def test_remove_duplicates__with_constant_key_is_one_item():
    assert remove_duplicates([(3, 2, 1), (3, 6, 5), (3, 6, 5)], key=lambda m: "asdf") == [(3, 2, 1)]

def test_remove_duplicates__with_no_key_uses_sorted_tuples():
    assert remove_duplicates([(3, 2, 1), (3, 6, 5), (3, 5, 6)]) == [(3, 2, 1), (3, 6, 5)]

def test_remove_duplicates__pick_random_removes_at_random():
    random.seed(1)
    assert remove_duplicates([(3, 2, 1), (3, 6, 5), (3, 5, 6)], pick_random=True) == [(3, 2, 1), (3, 6, 5)]
    random.seed(4)
    assert remove_duplicates([(3, 2, 1), (3, 6, 5), (3, 5, 6)], pick_random=True) == [(3, 2, 1), (3, 5, 6)]

def test_position_index_farthest_from_axis__for_octane_is_index_6_or_20(octane):
    # index 20 is one of the H on the benzene ring
    assert position_index_farthest_from_axis(octane.positions[-1] - octane.positions[0], octane) in [6,20]

def test_guess_elements_from_masses__finds_CNHHOC():
    assert guess_elements_from_masses([12.0, 14.0, 1.0, 1.0, 16.0, 12.0]) == ["C", "N", "H", "H", "O", "C"]

def test_guess_elements_from_masses__unknown_mass_raises_exception():
    with pytest.raises(Exception):
        guess_elements_from_masses([1700])




import csv
import sys
import time

import numpy as np

from mofun import Atoms, replace_pattern_in_structure, find_pattern_in_structure

# all_repldims = [(1,1,1), (2,2,2)]
# all_repldims = [(1,1,1), (2,2,2), (4,4,4)]
all_repldims = [(1,1,1), (2,2,2), (4,4,4), (5,5,5), (6,6,6), (7,7,7), (8,8,8)]

uio66 = Atoms.load("uio66.cif")
uio66_linker = Atoms.load("uio66-linker.cml")


output_csv = csv.writer(open("perf.csv", "w"))
output_csv.writerow(["mof", "repl", "operation", "num_atoms", "matches_found", "matches_expected", "process_time-seconds", "perf_counter_seconds"])


for repldims in all_repldims:
    print("find: ", repldims, )
    structure = uio66.replicate(repldims=repldims)
    time_s = time.process_time()
    perf_s = time.perf_counter()
    patterns = find_pattern_in_structure(structure, uio66_linker)
    output_csv.writerow(("uio66", "%dx%dx%dx" % repldims, "find", len(structure), len(patterns), 24 * repldims[0] * repldims[1] * repldims[2], time.process_time() - time_s, time.perf_counter() - perf_s))
    print(time.process_time() - time_s)

for repldims in all_repldims:
    print("replace: ", repldims)
    structure = uio66.replicate(repldims=repldims)
    time_s = time.process_time()
    perf_s = time.perf_counter()
    patterns = replace_pattern_in_structure(structure, uio66_linker, uio66_linker)
    output_csv.writerow(("uio66", "%dx%dx%dx" % repldims, "replace", len(structure), len(patterns), 24 * repldims[0] * repldims[1] * repldims[2], time.process_time() - time_s, time.perf_counter() - perf_s))
    print(time.process_time() - time_s)

## UIO-67

# uio67 = Atoms.load("uio67.cif")
# uio67_linker = Atoms.load("uio67-linker.cml")


# for repldims in all_repldims:
#     structure = uio67.replicate(repldims=repldims)
#     time_s = time.process_time()
#     perf_s = time.perf_counter()
#     patterns = find_pattern_in_structure(structure, uio67_linker)
#     output_csv.writerow(("uio67", "%dx%dx%dx" % repldims, len(structure), len(patterns), 24 * repldims[0] * repldims[1] * repldims[2], time.process_time() - time_s, time.perf_counter() - perf_s))

# for repldims in all_repldims:
#     structure = uio67.replicate(repldims=repldims)
#     time_s = time.process_time()
#     perf_s = time.perf_counter()
#     patterns = replace_pattern_in_structure(structure, uio67_linker, uio67_linker)
#     output_csv.writerow(("uio67", "%dx%dx%dx" % repldims, len(structure), len(patterns), time.process_time() - time_s, time.perf_counter() - perf_s))









from math import sqrt
from pathlib import Path

import ase.io
import numpy as np
from numpy.linalg import norm
import pytest
from pytest import approx

import tests
from mofun import Atoms
from mofun.helpers import typekey

sqrt2_2 = sqrt(2) / 2
sqrt3_2 = sqrt(3) / 2

def random_positions(num):
    return np.random.rand(num, 3) * 100

def assert_topo(topo, expected_topo, types=None, expected_types=None, coeffs=None, expected_coeffs=None):

    # check right atoms are part of the topo
    sorted_topo = sorted([typekey(t) for t in topo])
    sorted_expected_topo = sorted([typekey(t) for t in expected_topo])
    assert np.array_equal(sorted_topo, sorted_expected_topo)

    # check types are mapped (assume coeffs are ordered the same!)
    if types is not None and expected_types is not None:
        sorted_topo_w_types = sorted([(*typekey(t), types[i]) for i, t in enumerate(topo)])
        sorted_expected_topo_w_types = sorted([(*typekey(t), expected_types[i]) for i, t in enumerate(expected_topo)])
        assert np.array_equal(sorted_topo_w_types, sorted_expected_topo_w_types)

    # check coeffs for each type are equal
    if coeffs is not None and expected_coeffs is not None:
        assert np.array_equal(coeffs, expected_coeffs)

def assert_benzene(coords):
    # incomplete sample
    p = coords
    assert norm(p[0] - p[1]) == approx(2.42, 5e-2)
    assert norm(p[0] - p[3]) == approx(1.40, 5e-2)
    assert norm(p[0] - p[4]) == approx(2.79, 5e-2)
    assert norm(p[5] - p[8]) == approx(0.93, 5e-2)

@pytest.fixture
def linear_cnnc():
    yield Atoms(elements='CNNC', positions=[(0., 0., 0), (1.0, 0., 0.), (2.0, 0., 0.), (3.0, 0., 0.)],
                bonds=[(0,1), (1,2), (2,3)], bond_types=[0] * 3,
                angles=[(0,1,2), (1,2,3)], angle_types=[0,0],
                dihedrals=[(0,1,2,3)], dihedral_types=[0], cell=15*np.identity(3))

@pytest.fixture
def octane():
    # CH3 CH2 CH2 CH2 CH2 CH2 CH2 CH3 #
    with Path("tests/molecules/octane.xyz") as path:
        structure = Atoms.from_ase_atoms(ase.io.read(path))
        structure.cell = 60 * np.identity(3)
        structure.translate((30., 30., 30.))
        yield structure

@pytest.fixture
def half_octane():
    # CH3 CH2 CH2 CH2 #
    with Path("tests/molecules/half_octane.xyz") as path:
        structure = Atoms.from_ase_atoms(ase.io.read(path))
        structure.cell = 60 * np.identity(3)
        structure.translate((30., 30., 30.))
        yield structure


@pytest.fixture
def hkust1_cif():
    with Path("tests/hkust-1/hkust-1-with-bonds.cif") as path:
        yield Atoms.load_cif(path)

@pytest.fixture
def hkust1_3x3x3_xyz():
    with Path("tests/hkust-1/hkust-1-3x3x3.xyz") as path:
        structure = Atoms.from_ase_atoms(ase.io.read(path))
        structure.cell = 79.0290 * np.identity(3)
        yield structure

@pytest.fixture
def hkust1_3x3x3_cif():
    with Path("tests/hkust-1/hkust-1-3x3x3.cif") as path:
        yield Atoms.load_cif(path)

@pytest.fixture
def benzene():
    with Path("tests/molecules/benzene.xyz") as path:
        yield Atoms.from_ase_atoms(ase.io.read(path))

@pytest.fixture
def uio66_linker_no_bonds():
    with Path("tests/uio66/uio66-linker-no-bonds.lmpdat").open() as fd:
        yield Atoms.load_lmpdat(fd, atom_format="atomic")

@pytest.fixture
def uio66_linker_some_bonds():
    # this was a modified UIO-66-F linker with bonds defined for the C-F bond. The F's have been
    # replaced by H's.
    with Path("tests/uio66/uio66-linker.lmpdat").open() as fd:
        yield Atoms.load_lmpdat(fd, atom_format="atomic")

@pytest.fixture
def uio66_linker_cml():
    with Path("tests/uio66/uio66-linker.cml") as path:
        yield Atoms.load_cml(path)



import os

import networkx as nx

from mofun import Atoms
from mofun.rough_uff import calc_uff_atom_types, add_aromatic_flag, guess_bond_order, \
                            bond_params, angle_params, dihedral_params, calc_angles, calc_dihedrals, retype_atoms_from_uff_types
from tests.fixtures import uio66_linker_cml

from pytest import approx

def test_guess_bond_order__two_tetrahedrals_is_1():
    assert guess_bond_order('C_3', 'C_3') == 1
    assert guess_bond_order('N_3', 'N_3') == 1
    assert guess_bond_order('O_3', 'O_3') == 1
    assert guess_bond_order('O_3', 'C_3') == 1
    assert guess_bond_order('N_3', 'C_3') == 1
    assert guess_bond_order('N_3', 'O_3') == 1

def test_guess_bond_order__H_and_group_7_elements_is_1():
    assert guess_bond_order('H_', 'C_3') == 1
    assert guess_bond_order('F_', 'C_3') == 1
    assert guess_bond_order('Cl', 'C_3') == 1
    assert guess_bond_order('Br', 'C_3') == 1
    assert guess_bond_order('I', 'C_3') == 1

def test_guess_bond_order__CNO_trigonal_pairs_is_2():
    assert guess_bond_order('C_2', 'C_2') == 2
    assert guess_bond_order('N_2', 'N_2') == 2
    assert guess_bond_order('O_2', 'O_2') == 2
    assert guess_bond_order('C_2', 'N_2') == 1
    assert guess_bond_order('N_2', 'O_2') == 1
    assert guess_bond_order('O_2', 'C_2') == 1

def test_guess_bond_order__CNO_aromatic_pairs_is_1_5():
    assert guess_bond_order('C_R', 'C_R') == 1.5
    assert guess_bond_order('N_R', 'N_R') == 1.5
    assert guess_bond_order('O_R', 'O_R') == 1.5
    assert guess_bond_order('C_R', 'N_2') == 1
    assert guess_bond_order('N_R', 'O_2') == 1
    assert guess_bond_order('O_R', 'C_2') == 1

def test_guess_bond_order__rules_for_azido_is_2():
    rules = [({'N_1'}, 2), ({'N_1', 'N_2'}, 2)]
    assert guess_bond_order('N_1', 'N_1') == 1
    assert guess_bond_order('N_1', 'N_1', rules) == 2
    assert guess_bond_order('N_1', 'N_2') == 1
    assert guess_bond_order('N_1', 'N_2', rules) == 2

def test_calc_uff_atom_types__carbon_in_methane_is_tetrahedral():
    uff_atom_types = calc_uff_atom_types([(0,1), (0,2), (0,3), (0,4)], ["C", "H", "H", "H", "H"])
    assert uff_atom_types == ["C_3", "H_", "H_", "H_", "H_"]

def test_calc_uff_atom_types__carbon_in_co2_is_linear():
    uff_atom_types = calc_uff_atom_types([(0,1), (0,2)], ["C", "O", "O"])
    assert uff_atom_types == ["C_1", "O_1", "O_1"]

def test_calc_uff_atom_types__carbon_in_benzene_ring_is_aromatic():
    uff_atom_types = calc_uff_atom_types([(0,1), (1,2), (2,3), (3,4), (4,5), (5,0)], ["C"] * 6)
    assert uff_atom_types == ["C_R"] * 6

def test_calc_uff_atom_types__N_in_RNH2_is_tetrahedral():
    uff_atom_types = calc_uff_atom_types([(0,1), (0,2), (0,3), (1,4)], ["N", "C", "H", "H", "H"])
    assert uff_atom_types[0] == "N_3"

def test_calc_uff_atom_types__N_in_RNNN_is_trigonal():
    uff_atom_types = calc_uff_atom_types([(0,1), (1,2), (0,3), (3,4)], ["N", "N", "N", "C", "H"])
    assert uff_atom_types[0] == "N_2"

def test_calc_uff_atom_types__N_in_RNC4_ring_is_aromatic():
    uff_atom_types = calc_uff_atom_types([(0,1), (1,2), (2,3), (3,4), (4,0), (0,5), (5,6)], ["N", "C", "C", "C", "C", "C", "H"])
    assert uff_atom_types[0] == "N_R"

def test_calc_uff_atom_types__NNN_override_rule_neighbors_assigns_N1():
    override_rules = {
        "N": [
            ("N_1", dict(neighbors=("N","N"))),
        ]
    }
    uff_atom_types = calc_uff_atom_types([(0,1), (1,2), (0,3), (3,4)], ["N", "N", "N", "C", "H"], override_rules=override_rules)
    assert uff_atom_types[0:3] == ["N_2", "N_1", "N_1"]

def test_calc_uff_atom_types__elements_with_only_one_uff_atom_type_wo_hybridization_get_that_type():
    uff_atom_types = calc_uff_atom_types([(0,1), (1,2), (2,3), (3,4), (4,5), (5,0)], ["F", "Rb", "Li", "Cl", "K", "Br"])
    assert uff_atom_types == ["F_", "Rb", "Li", "Cl", "K_", "Br"]

def test_add_aromatic_flag__uio66_linker_has_aromatic_benzene_ring(uio66_linker_cml):
    g = nx.Graph()
    g.add_edges_from(uio66_linker_cml.bonds)
    add_aromatic_flag(g)
    atom_benzene_ring = ["aromatic" in g.nodes[n] for n in sorted(g.nodes())]
    # Trues indicate where the benzene carbons are in the uio_linker_cml file
    assert atom_benzene_ring == [False, False, False, True, True, False, False, True, True, True, True, False, False, False, False, False]

def test_calc_uff_atom_types__uio66_linker(uio66_linker_cml):
    uff_atom_types = calc_uff_atom_types(uio66_linker_cml.bonds, uio66_linker_cml.elements)
    assert uff_atom_types == ["C_2", "O_1", "O_1", "C_R", "C_R", "H_", "H_", "C_R", "C_R", "C_R",
                              "C_R", "H_", "H_", "O_1", "C_2", "O_1"]

def test_bond_params__C_N_amide_is_1293():
    # normative values for bond length and force constant from Towhee
    force_k, bond_length = bond_params("C_R", "N_R", bond_order=1.41)
    assert bond_length == approx(1.3568, 1e-4)
    # /2 here because we are currently returning constants compatible with LAMMPS
    assert force_k == approx(1293.18/2, 5e-2)

def test_angle_params__C_N_C_amide_is_105_5():
    # normative values for force constant from Towhee
    angle_style, force_k, b, n = angle_params("C_3", "N_R", "C_R", bond_orders=[None, 1.41])
    print(angle_style, force_k, b, n)
    assert angle_style  == "cosine/periodic"
    assert force_k == approx(210.97397, 1e-5)

def test_dihedral_params__force_constant_should_match_table_2_kind_of():
    # Note that none of the calculated force constants match table 2 because the dihderal potential
    # is ill-defined. The normative values here come from our attempt at calculating what they
    # should be.

    # test SP3-SP3
    assert dihedral_params("H_", "C_3", "C_3", "H_")[1] * 2 == approx(2.119, abs=1e-3)
    assert dihedral_params("H_", "C_3", "Si3", "H_")[1] * 2 == approx(1.611, abs=1e-3)
    assert dihedral_params("H_", "C_3", "Ge3", "H_")[1] * 2 == approx(1.219, abs=1e-3)
    assert dihedral_params("H_", "C_3", "Sn3", "H_")[1] * 2 == approx(0.649, abs=1e-3)
    assert dihedral_params("H_", "C_3", "N_3", "H_")[1] * 2 == approx(0.977, abs=1e-3)
    assert dihedral_params("H_", "C_3", "P_3+3", "H_")[1] * 2 == approx(2.255, abs=1e-3)
    assert dihedral_params("H_", "C_3", "As3+3", "H_")[1] * 2 == approx(1.783, abs=1e-3)
    assert dihedral_params("H_", "C_3", "O_3", "H_")[1] * 2 == approx(0.195, abs=1e-3)
    assert dihedral_params("H_", "C_3", "S_3+2", "H_")[1] * 2 == approx(1.013, abs=1e-3)
    assert dihedral_params("H_", "C_3", "Se3+2", "H_")[1] * 2 == approx(0.843, abs=1e-3)

    # test SP3-SP3 exception for oxygen group elements
    assert dihedral_params("H_", "O_3", "O_3", "H_")[1] * 2 == approx(2.0, abs=1e-3)
    assert dihedral_params("H_", "S_3+2", "S_3+2", "H_")[1] * 2 == approx(6.8, abs=1e-3)

    # test SP2-SP2
    assert dihedral_params("C_", "O_2", "C_2", "C_")[1] * 2 == approx(10., abs=1e-3)
    assert dihedral_params("C_", "O_2", "S_2", "C_")[1] * 2 == approx(7.906, abs=1e-3)

    # test SP3-SP2
    # propene
    assert dihedral_params("H_", "C_3", "C_2", "C_2")[1] * 2 == approx(2., abs=1e-3)
    assert dihedral_params("C_2", "C_2", "C_3", "H_")[1] * 2 == approx(2., abs=1e-3)

def test_dihedral_params__X_1_is_none():
    assert dihedral_params('C_R', 'N_2', 'N_1', 'N_1', 1) is None

def test_calc_angles__ethane_has_12_angles():
    angles = calc_angles([(0,3), (1,3), (2,3), (3,4), (4,5), (4,6), (4,7)])
    unique_angles = set([tuple(x) for x in angles])
    assert len(angles) == 12
    assert unique_angles == {(0,3,1), (0,3,2), (1,3,2), (0,3,4), (1,3,4), (2,3,4),
                             (3,4,5), (3,4,6), (3,4,7), (5,4,6), (5,4,7), (6,4,7)}

def test__dihedrals__ethane_has_9_dihedrals():
    dihedrals = calc_dihedrals([(0,3), (1,3), (2,3), (3,4), (4,5), (4,6), (4,7)])
    unique_dihedrals = set([tuple(x) for x in dihedrals])
    assert len(dihedrals) == 9
    assert unique_dihedrals == {(0,3,4,5), (1,3,4,5), (2,3,4,5),
                                (0,3,4,6), (1,3,4,6), (2,3,4,6),
                                (0,3,4,7), (1,3,4,7), (2,3,4,7)}

def test__retype_atoms_from_uff_types__boron_gives_all_boron():
    cnnc = Atoms(elements='CNNC', positions=[(0., 0., 0), (1.0, 0., 0.), (2.0, 0., 0.), (3.0, 0., 0.)])
    retype_atoms_from_uff_types(cnnc, ["B"] * 4)
    assert cnnc.atom_types == [0] * 4
    assert cnnc.atom_type_labels == ["B"]
    assert cnnc.atom_type_elements == ["B"]
    assert cnnc.atom_type_masses == [10.811]

def test__retype_atoms_from_uff_types__BZrZrB_gives_BZrZrB():
    cnnc = Atoms(elements='CNNC', positions=[(0., 0., 0), (1.0, 0., 0.), (2.0, 0., 0.), (3.0, 0., 0.)])
    retype_atoms_from_uff_types(cnnc, ["B", "Zr", "Zr", "B"])
    assert cnnc.atom_types == [0, 1, 1, 0]
    assert cnnc.atom_type_labels == ["B", "Zr"]
    assert cnnc.atom_type_elements == ["B", "Zr"]
    assert cnnc.atom_type_masses == [10.811, 91.224]

def test__retype_atoms_from_uff_types__BZrZrB_with_full_atom_types_gives_BZrZrB():
    cnnc = Atoms(elements='CNNC', positions=[(0., 0., 0), (1.0, 0., 0.), (2.0, 0., 0.), (3.0, 0., 0.)])
    retype_atoms_from_uff_types(cnnc, ["B_3", "Zr8f4", "Zr8f4", "B_3"])
    assert cnnc.atom_types == [0, 1, 1, 0]
    assert cnnc.atom_type_labels == ["B_3", "Zr8f4"]
    assert cnnc.atom_type_elements == ["B", "Zr"]
    assert cnnc.atom_type_masses == [10.811, 91.224]

def test__retype_atoms_from_uff_types__C2C1ZrC3H_with_full_atom_types_orders_as_HCCCZr():
    cnnnc = Atoms(elements='CNNNC', positions=[(0., 0., 0), (1.0, 0., 0.), (2.0, 0., 0.), (3.0, 0., 0.), (4.0, 0., 0.)])
    retype_atoms_from_uff_types(cnnnc, ["C_2", "C_1", "Zr8f4", "C_3", "H_"])
    assert cnnnc.atom_types == [2, 1, 4, 3, 0]
    assert cnnnc.atom_type_labels == ["H_", "C_1", "C_2", "C_3", "Zr8f4"]
    assert cnnnc.atom_type_elements == ["H", "C", "C", "C", "Zr"]
    assert cnnnc.atom_type_masses == [1.00794, 12.0107, 12.0107, 12.0107, 91.224]




import click
import matplotlib.pyplot as plt
from matplotlib import rc
import numpy as np
import pandas as pd

rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})

fsl = fs = 9

@click.command()
@click.argument('csv-path', type=click.File())
@click.option('--outputpath', '-o', type=click.Path(), default="perf-plot.png")
def perfplot(csv_path, outputpath="perf-plot.png"):
    fig = plt.figure(figsize=(3.3, 3.3))

    data = pd.read_csv(csv_path)
    find = data[data['operation'] == 'find']
    replace = data[data['operation'] == 'replace']
    print(data)

    ax = fig.subplots(ncols=1)
    ax.set_xticks([0, 50000, 100000, 150000, 200000, 250000])
    ax.set_xticklabels(["0", "50K", "100K", "150K", "200K", "250K"])

    ax.set_ylim((-10, 400))
    ax.set_xlim((-15000, 250000))

    ax.grid(which='major', axis='x', linestyle='-', color='0.85', zorder=10)
    ax.grid(which='major', axis='y', linestyle='-', color='0.85', zorder=10)

    ax.set_ylabel("Time (s)")
    ax.set_xlabel("# Atoms")

    ax.scatter(pd.to_numeric(find['num_atoms']), find['process_time_seconds'], label="Find", zorder=50)
    ax.scatter(pd.to_numeric(replace['num_atoms']), replace['process_time_seconds'], label="Find + Replace", zorder=50)

    for row in replace.itertuples():
        print(row)
        ax.annotate(row.repl, (row.num_atoms, row.process_time_seconds), xytext=(0,3), textcoords='offset points', horizontalalignment="center", verticalalignment='bottom', fontsize=7.5, c="black")

    ax.legend()

    fig.savefig(outputpath, dpi=300, bbox_inches='tight')#, transparent=True)
    plt.close(fig)


if __name__ == '__main__':
    perfplot()




import io
from io import StringIO

from tests.fixtures import *
from mofun import Atoms
from mofun.atoms import find_unchanged_atom_pairs
from mofun.helpers import assert_positions_are_unchanged, assert_structure_positions_are_unchanged


def test_atoms__type_labels_are_inferred_from_elements():
    atoms = Atoms(elements="CNZrPt", positions=random_positions(4))
    assert np.array_equal(atoms.atom_type_labels, ["C", "N", "Zr", "Pt"])

def test_atoms__masses_are_inferred_from_elements():
    atoms = Atoms(elements="CNZrPt", positions=random_positions(4))
    assert np.allclose(atoms.atom_type_masses, [12, 14, 91.2, 195.1], atol=0.1)

def test__delete_and_reindex_atom_index_array():
    a = np.array([[1,2], [3,4], [5,6], [1,6]])
    updated_atoms = Atoms()._delete_and_reindex_atom_index_array(a, [3])
    assert(updated_atoms == np.array([[1,2],[4,5],[1,5]])).all()

def test_atoms_del__deletes_bonds_attached_to_atoms(linear_cnnc):
    del(linear_cnnc[[1]])
    assert list(linear_cnnc.elements) == ["C", "N", "C"]
    assert (linear_cnnc.bonds == [[1,2]]).all()

def test_atoms_del__deletes_types_with_all_topologies(linear_cnnc):
    linear_cnnc.impropers = np.array([(0,1,2,3)])
    linear_cnnc.bond_types = np.array([0, 1, 2])
    linear_cnnc.angle_types = np.array([0, 1])
    linear_cnnc.dihedral_types = np.array([0])
    linear_cnnc.improper_types = np.array([0])
    del(linear_cnnc[[0]])
    assert (linear_cnnc.bond_types == (1, 2)).all()
    assert linear_cnnc.angle_types == (1)
    assert len(linear_cnnc.dihedral_types) == 0
    assert len(linear_cnnc.improper_types) == 0

def test_atoms_del__deletes_angles_attached_to_atoms(linear_cnnc):
    del(linear_cnnc[[0]])
    assert (linear_cnnc.angles == [(0,1,2)]).all()

def test_atoms_del__deletes_dihedrals_attached_to_atoms(linear_cnnc):
    del(linear_cnnc[[1]])
    assert len(linear_cnnc.dihedrals) == 0

def test_atoms_del__deletes_impropers_attached_to_atoms():
    atoms = Atoms(elements="HCHH", positions=random_positions(4), impropers=[(0,1,2,3)], improper_types=[0])
    del(atoms[[1]])
    assert len(atoms.impropers) == 0

def test_atoms_extend__on_nonbonded_structure_reindexes_new_bonds_to_proper_atoms(linear_cnnc):
    linear_cnnc_no_bonds = Atoms(elements='CNNC', positions=[(0., 0., 0), (1.0, 0., 0.), (2.0, 0., 0.), (3.0, 0., 0.)])
    linear_cnnc_no_bonds.extend(linear_cnnc)
    assert np.array_equal(linear_cnnc_no_bonds.bonds, [(4,5), (5,6), (6,7)])

def test_atoms_extend__new_types_come_after_old_types1(linear_cnnc):
    a = Atoms(elements="C", positions=[[0, 0, 0]])
    b = Atoms(elements="H", positions=[[1, 1, 1]])
    a.extend(b)
    assert np.array_equal(a.elements, ["C", "H"])

def test_atoms_extend__new_types_come_after_old_types(linear_cnnc):
    linear_cnnc.impropers = np.array([(0,1,2,3)])

    linear_cnnc.atom_types = np.array([0,1,1,0])
    linear_cnnc.bond_types = np.array([0,1,0])
    linear_cnnc.angle_types = np.array([0,1])
    linear_cnnc.dihedral_types = np.array([0])
    linear_cnnc.improper_types = np.array([0])
    double_cnnc = linear_cnnc.copy()
    double_cnnc.extend(linear_cnnc)
    assert np.array_equal(double_cnnc.atom_types, [0, 1, 1, 0, 2, 3, 3, 2])
    assert np.array_equal(double_cnnc.bond_types, [0, 1, 0, 2, 3, 2])
    assert np.array_equal(double_cnnc.angle_types, [0, 1, 2, 3])
    assert np.array_equal(double_cnnc.dihedral_types, [0, 1])
    assert np.array_equal(double_cnnc.improper_types, [0, 1])
    assert np.array_equal(double_cnnc.elements, ["C", "N", "N", "C"] * 2)

def test_atoms_extend__with_structure_map_reindexes_new_bonds_to_proper_atoms(linear_cnnc):
    fn_pattern = Atoms(elements='NCH', atom_type_labels=["Nx", "Cx", "Hx"],
                positions=[(2.0, 0., 0.), (3.0, 0., 0.), (4.0, 0., 0.)],
                bonds=[(0,1), (1,2)], bond_types=[0] * 2,
                angles=[(0,1,2)], angle_types=[0])
    fn_linear_cnnc = linear_cnnc.copy()
    fn_linear_cnnc.extend(fn_pattern, structure_index_map={0:2, 1:3})

    assert len(fn_linear_cnnc) == 5
    assert np.array_equal(fn_linear_cnnc.bonds, [[0,1], [1,2], [2,3], [3,4]])
    assert np.array_equal(fn_linear_cnnc.angles, [[0,1,2], [1,2,3], [2,3,4]])
    assert np.array_equal(fn_linear_cnnc.atom_types, [0, 1, 2, 3, 4])
    assert np.array_equal(fn_linear_cnnc.atom_type_labels, ["C", "N", "Nx", "Cx", "Hx"])

def test_atoms_extend__reindexes_new_bonds_to_proper_atoms(linear_cnnc):
    double_cnnc = linear_cnnc.copy()
    double_cnnc.extend(linear_cnnc)
    assert np.array_equal(double_cnnc.bonds, [(0,1), (1,2), (2,3), (4,5), (5,6), (6,7)])

def test_atoms_extend__reindexes_new_angles_to_proper_atoms(linear_cnnc):
    double_cnnc = linear_cnnc.copy()
    double_cnnc.extend(linear_cnnc)
    assert (double_cnnc.angles == [(0,1,2), (1,2,3), (4,5,6), (5,6,7)]).all()

def test_atoms_extend__reindexes_new_dihedrals_to_proper_atoms(linear_cnnc):
    double_cnnc = linear_cnnc.copy()
    double_cnnc.extend(linear_cnnc)
    assert (double_cnnc.dihedrals == [(0,1,2,3), (4,5,6,7)]).all()

def test_atoms_extend__reindexes_new_impropers_to_proper_atoms(linear_cnnc):
    linear_cnnc.impropers = np.array([(0,1,2,3)])
    linear_cnnc.improper_types = np.array([0])
    double_cnnc = linear_cnnc.copy()
    double_cnnc.extend(linear_cnnc)
    assert (double_cnnc.impropers == [(0,1,2,3), (4,5,6,7)]).all()


def test_find_unchanged_atom_pairs__same_structure_is_unchanged(linear_cnnc):
    assert find_unchanged_atom_pairs(linear_cnnc, linear_cnnc) == [(0,0), (1,1), (2,2), (3,3)]

def test_find_unchanged_atom_pairs__different_atom_type_is_changed(linear_cnnc):
    cncc = linear_cnnc.copy()
    cncc.atom_types[2] = cncc.atom_type_elements.index("C")
    assert find_unchanged_atom_pairs(linear_cnnc, cncc) == [(0,0), (1,1), (3,3)]

def test_find_unchanged_atom_pairs__different_position_is_changed(linear_cnnc):
    offset_linear_cnns = linear_cnnc.copy()
    offset_linear_cnns.positions[2] += 0.5
    assert find_unchanged_atom_pairs(linear_cnnc, offset_linear_cnns) == [(0,0), (1,1), (3,3)]

def test_atoms_getitem__has_all_atom_types_and_charges():
    atoms = Atoms(atom_type_elements=["C", "N"], atom_types=[1, 0, 0, 1], positions=[[0,0,0]] * 4, charges=[1,2,3,4])
    assert atoms[0].elements[0] == "N"
    assert atoms[1].elements[0] == "C"
    assert atoms[0].charges[0] == 1
    assert atoms[1].charges[0] == 2
    assert atoms[(0,1)].elements == ["N", "C"]
    assert atoms[(2,3)].elements == ["C", "N"]
    assert (atoms[(0,1)].charges ==[1, 2]).all()

def test_atoms_replicate__111_is_unchanged(octane):
    reploctane = octane.replicate((1,1,1))
    assert np.array_equal(octane.positions, reploctane.positions)
    assert np.array_equal(octane.atom_types, reploctane.atom_types)
    assert np.array_equal(octane.charges, reploctane.charges)
    assert np.array_equal(octane.groups, reploctane.groups)

def test_atoms_replicate__211_has_replicate_in_x_dim(octane):
    reploctane = octane.replicate((2,1,1))
    assert (octane.positions == reploctane.positions[0:26, :]).all()
    assert (octane.positions[:, 0] + 60 == reploctane.positions[26:52, 0]).all()
    assert (octane.positions[:, 1:] == reploctane.positions[26:52, 1:]).all()

    assert np.array_equal(np.tile(octane.atom_types,2), reploctane.atom_types)
    assert np.array_equal(np.tile(octane.charges,2), reploctane.charges)
    assert np.array_equal(np.tile(octane.groups,2), reploctane.groups)

def test_atoms_replicate__213_has_replicates_in_xz_dims(octane):
    reploctane = octane.replicate((2,1,3))
    na = 26 # atoms in octane

    # check [1, 1, 1]
    assert (octane.positions == reploctane.positions[0:26, :]).all()

    # check [2, 1, 1]
    i = 1
    assert (octane.positions[:, 0] + 60 == reploctane.positions[na*i:na*(i+1), 0]).all()
    assert (octane.positions[:, 1] +  0 == reploctane.positions[na*i:na*(i+1), 1]).all()
    assert (octane.positions[:, 2] +  0 == reploctane.positions[na*i:na*(i+1), 2]).all()

    # check [1, 1, 2]
    i = 2
    assert (octane.positions[:, 0] +  0 == reploctane.positions[na*i:na*(i+1), 0]).all()
    assert (octane.positions[:, 1] +  0 == reploctane.positions[na*i:na*(i+1), 1]).all()
    assert (octane.positions[:, 2] + 60 == reploctane.positions[na*i:na*(i+1), 2]).all()

    # check [2, 1, 2]
    i = 3
    assert (octane.positions[:, 0] + 60 == reploctane.positions[na*i:na*(i+1), 0]).all()
    assert (octane.positions[:, 1] +  0 == reploctane.positions[na*i:na*(i+1), 1]).all()
    assert (octane.positions[:, 2] + 60 == reploctane.positions[na*i:na*(i+1), 2]).all()

    # check [1, 1, 3]
    i = 4
    assert (octane.positions[:, 0] +   0 == reploctane.positions[na*i:na*(i+1), 0]).all()
    assert (octane.positions[:, 1] +   0 == reploctane.positions[na*i:na*(i+1), 1]).all()
    assert (octane.positions[:, 2] + 120 == reploctane.positions[na*i:na*(i+1), 2]).all()

    # check [2, 1, 3]
    i = 5
    assert (octane.positions[:, 0] +  60 == reploctane.positions[na*i:na*(i+1), 0]).all()
    assert (octane.positions[:, 1] +   0 == reploctane.positions[na*i:na*(i+1), 1]).all()
    assert (octane.positions[:, 2] + 120 == reploctane.positions[na*i:na*(i+1), 2]).all()

    assert np.array_equal(np.tile(octane.atom_types, 6), reploctane.atom_types)
    assert np.array_equal(np.tile(octane.charges, 6), reploctane.charges)
    assert np.array_equal(np.tile(octane.groups, 6), reploctane.groups)

def test_atoms_replicate__triclinic_222_has_replicates_in_three_dims():
    a = Atoms(elements="H", positions=[(1, 1, 1)], cell=np.array([[10, 0, 0], [10, 10, 0], [0, 0, 10]]))
    expected = Atoms(elements="HHHHHHHH", cell=np.array([[20, 0, 0], [20, 20, 0], [0, 0, 20]]),
                     positions=[[1,1,1], [11,1,1], [11,11,1], [21,11,1],
                                [1,1,11], [11,1,11], [11,11,11], [21,11,11]])

    ra = a.replicate((2, 2, 2))
    assert len(ra) == 8
    assert_structure_positions_are_unchanged(ra, expected)
    assert np.array_equal(ra.cell, expected.cell)



import copy
import io
import os
import pathlib
import xml.etree.ElementTree as ET

import ase
from ase.formula import Formula
from CifFile import ReadCif as read_cif

import numpy as np
from scipy.linalg import norm
from scipy.spatial.distance import cdist

from mofun.helpers import guess_elements_from_masses, ATOMIC_MASSES, use_or_open

class Atoms:
    """
    An Atoms object is a container for all the information required to keep track of a structure and
    the structure's force field.

    See the documentation on `__init__` or `load` to see how to create an Atoms object.

    An Atoms object is made up primarily of numpy arrays and a couple normal lists.

    Some arrays are per-atom (one entry per atom): `atom_types`, `positions`, `charges`, and
    `groups`.

    Some arrays are per-atom-type (one entry per atom type): `atom_type_masses`, `atom_type_elements`, and
    `atom_type_labels`.

    Some arrays are per-bond (one entry per bond): `bonds`, and `bond_types`.

    Some arrays are per-angle (one entry per angle): `angles`, and `angle_types`.

    Some arrays are per-dihedral (one entry per dihedral): `dihedrals`, and `dihedral_types`.

    Some arrays are per-improper (one entry per improper): `impropers`, and `improper_types`.

    Atoms also stores information on the coefficients needed to define each force field term:
    `pair_coeffs`, `bond_type_coeffs`, `angle_type_coeffs`, `dihedral_type_coeffs`, and
    `improper_type_coeffs`. The \\*\\_coeffs variables are lists of strings, where the item index
    corresponds to the \\*\\_type, and the string is the full LAMMPS coeffs definition string. we do
    not interpret any of the LAMMPS coefficient specifics, we just store it in its original form,
    i.e. this Angle Coeffs section:

    ```
    Angle Coeffs

    1 cosine/periodic  72.500283  -1  1   # C_R O_1 H_
    2 cosine/periodic  277.164705  -1  3   # C_R C_R O_1
    ```

    would be interpreted like this:

    ```
    angle_type_coeffs= ["cosine/periodic  72.500283  -1  1   # C_R O_1 H_",
                        "cosine/periodic  277.164705  -1  3   # C_R C_R O_1"]
    ```"""

    def __init__(self, atom_types=[], positions=[], charges=[], groups=[],
                    elements=[], atom_type_masses=[], atom_type_elements=[], atom_type_labels=[],
                    bonds=[], bond_types=[], angles=[], angle_types=[],
                    dihedrals=[], dihedral_types=[], impropers=[], improper_types=[],
                    pair_coeffs=[], bond_type_coeffs=[], angle_type_coeffs=[],
                    dihedral_type_coeffs=[], improper_type_coeffs=[], cell=None):
        """Create an Atoms object.

        An Atoms object can be created without any atoms using `Atoms()`. For creating more
        interesting Atoms objects, there are a few rules to keep in mind. The parameters
        `atom_types`, `positions`, `charges`, and `groups`  are all lists that should have a
        size equal to the number of atoms in the system. `positions` is mandatory; `charges`, and
        `groups` are optional (both default to 0 for each atom) and there are two ways to
        specify atom types: 1) specify the atom_types and atom_type_elements explicitly, which is
        how the `load_lmpdat` method loads Atoms objects from a LAMMPS data file, or 2) specify
        per-atom elements and and have MOFUN auto-number the atom types for you, which is more
        convenient when specifying small molecules in code or when loading from other file formats
        such as CML or CIF which may store element information but not type information.

        When explicitly setting the types, you must pass `atom_types` and `atom_type_elements`.
        `atom_types` is a list of int type ids >= 0, one for each atom in the system.
        `atom_type_elements` is a list of element names (e.g. "C", "N", "Zr") per _atom type_. For
        example, if your system is propane, your atom_types list could be [0, 1, 1, 1] and your
        atom_type_elements list would then be ["C", "H"].

        To specify per-atom elements, you must pass `elements` with either a list of elements, such
        as `Atoms(elements=["C", "C"], ...)` or with a string `Atoms(elements="CC", ...)`. If you
        use the `elements` parameter, then type ids are automatically generated.

        Passing `atom_type_masses` is optional, and masses will be inferred from the elements if
        missing. If you are using atom types with masses that do not correspond to periodic table
        elements, then you will need to specify the masses explicitly.

        Passing force field term information for `bonds`, `angles`, `dihedrals`, and `impropers` is optional, as well
        as passing force field coefficients for LAMMPS with `pair_coeffs`, `bond_type_coeffs`,
        `angle_type_coeffs`, `dihedral_type_coeffs`, and `improper_type_coeffs`.

        Examples:

        ```
         a = Atoms() # create an empty Atoms object with no atoms
         a = Atoms(atom_types=[0], positions=[[0,0,0]]) # create one atom of type 0
         a = Atoms(elements=["C"], positions=[[0,0,0]]) # create one Carbon
         a = Atoms(elements=["C", "C"], positions=[[0,0,0], [1,0,0]]) # create two Carbons
         a = Atoms(elements="CC", positions=[[0,0,0], [1,0,0]]) # create two Carbons using shorthand element notation
        ```

        Args:
            atom_types (List[int]): list of integer type ids, one per atom.
            positions (List[Tuple[float, float, float]]): list of tuple atom x,y,z coordinates, one per atom.
            charges (List[float]): list of charges, one per atom. Defaults to 0 for each atom if not passed.
            groups (List[int]): list of integer groups, one per atom. For LAMMPS this gets mapped to a "molecule id". Defaults to 0 for each atom if not passed.
            elements (List[str], str): either a list of elements, (e.g. ["C", "H", "H", "H"]) or an element string (e.g. "CHHH")
            atom_type_masses (List[float]): list of atom type masses, one per atom type. If masses are not passed, they will be inferred from the `atom_type_elements`.
            atom_type_elements (List[str]): list of atom type elements, one per atom type.
            atom_type_labels (List[str]):  list of atom type labels, one per atom type. Used in LAMMPS data file line comments as type labels.
            bonds (List[Tuple[int, int]]): list of bond tuples where each tuple defines a pair of atoms that are bonded. Each value in the tuple is an index of an atom in the atom_* lists.
            bond_types (List[int]): list of bond type id ints for each bond defined in `bonds`.
            angles (List[Tuple[int, int, int]]): list of angle tuples where each tuple defines a triplet of atoms making up the angle. Each value in the tuple is an index of an atom in the atom_* lists.
            angle_types (List[int]): list of angle type id ints for each angle defined in `angles`.
            dihedrals (List[Tuple[int, int, int, int]]): list of dihedral tuples where each dihedral defines a quartet of atoms making up the dihedral. Each value in the tuple is an index of an atom in the atom_* lists.
            dihedral_types (List[int]): list of dihedral type id ints for each dihedral defined in `dihedrals`.
            impropers (List[Tuple[int, int, int, int]]): list of improper tuples where each improper defines a quartet of atoms making up the improper. Each value in the tuple is an index of an atom in the atom_* lists.
            improper_types (List[int]): list of improper type id ints for each improper defined in `impropers`.
            pair_coeffs (List[str]): pair coefficients definition string (anything supported by LAMMPS in a data file but without the type id). One per atom type.
            bond_type_coeffs (List[str]): bond coefficients definition string (anything supported by LAMMPS in a data file but without the type id). One per bond type.
            angle_type_coeffs (List[str]): angle coefficients definition string (anything supported by LAMMPS in a data file but without the type id). One per angle type.
            dihedral_type_coeffs (List[str]): dihedral coefficients definition string (anything supported by LAMMPS in a data file but without the type id). One per dihedral type.
            improper_type_coeffs (List[str]): improper coefficients definition string (anything supported by LAMMPS in a data file but without the type id). One per improper type.
            cell (Array(3x3)): 3x3 array of unit cell vectors.

        Returns:
            Atoms: the atoms object.
        """

        self.atom_type_masses = np.array(atom_type_masses, ndmin=1)
        self.positions = np.array(positions, dtype=float, ndmin=1)

        if cell is not None:
            self.cell = np.array(cell)
        else:
            self.cell = None

        self.bonds = np.array(bonds, dtype=int)
        self.bond_types = np.array(bond_types, dtype=int)
        self.angles = np.array(angles, dtype=int)
        self.angle_types = np.array(angle_types, dtype=int)
        self.dihedrals = np.array(dihedrals, dtype=int)
        self.dihedral_types = np.array(dihedral_types, dtype=int)
        self.impropers = np.array(impropers, dtype=int)
        self.improper_types = np.array(improper_types, dtype=int)

        self.pair_coeffs = np.array(pair_coeffs)
        self.bond_type_coeffs = np.array(bond_type_coeffs)
        self.angle_type_coeffs = np.array(angle_type_coeffs)
        self.dihedral_type_coeffs = np.array(dihedral_type_coeffs)
        self.improper_type_coeffs = np.array(improper_type_coeffs)

        if len(charges) > 0:
            self.charges = np.array(charges, dtype=float)
        else:
            self.charges = np.zeros(len(self.positions), dtype=float)

        if len(groups) > 0:
            self.groups = np.array(groups, dtype=int)
        else:
            self.groups = np.zeros(len(self.positions), dtype=int)

        # load atom_types and atom_type_elements
        if len(atom_types) > 0:
            # default case or a __getitem__ subset
            self.atom_types = np.array(atom_types)
            self.atom_type_elements = atom_type_elements
        elif len(elements) > 0:
            # from element array, such as from ASE atoms or read CML
            # i.e. Propane ['C', 'H', 'H', 'H', 'C', 'H', 'H', 'C', 'H', 'H', 'H']:
            # or from element string, i.e. Propane "CHHHCHHCHHH" (shorthand):
            if isinstance(elements, str):
                elements = list(Formula(elements))

            # preserve order of types
            self.atom_type_elements = list(dict.fromkeys(elements).keys())
            self.atom_types = np.array([self.atom_type_elements.index(s) for s in elements])
        else:
            # no atom_type_elements or elements passed
            # this should be the `Atoms()` case; if not, it will fail the asserts below
            self.atom_types = np.array([], ndmin=1)
            self.atom_type_elements = []

        # automatically determine masses from elements if masses are not passed
        if len(self.atom_type_masses) == 0 and len(self.atom_type_elements) > 0:
            self.atom_type_masses = [ATOMIC_MASSES[s] for s in self.atom_type_elements]

        if len(atom_type_labels) > 0:
            self.atom_type_labels = atom_type_labels
        else:
            print("WARNING: using the atom elements as the atom_type_labels since labels were not supplied.")
            # use default atom types equal to the element; this may not be unique!
            self.atom_type_labels = self.atom_type_elements

        self.assert_arrays_are_consistent_sizes()

    def assert_arrays_are_consistent_sizes(self):
        # we always should have these lists, though charges and groups may be zero defaults
        if len(self.positions) != len(self.atom_types):
            raise Exception("len of positions (%d) and atom types (%d) must match" % (len(self.positions), len(self.atom_types)))
        if len(self.positions) != len(self.charges):
            raise Exception("len of positions (%d) and charges (%d) must match" % (len(self.positions), len(self.charges)))
        if len(self.positions) != len(self.groups):
            raise Exception("len of positions (%d) and groups (%d) must match" % (len(self.positions), len(self.groups)))

        # these list pairs should always match
        if len(self.bonds) != len(self.bond_types):
            raise Exception("len of bonds and bond types must match")
        if len(self.angles) != len(self.angle_types):
            raise Exception("len of angles and angle types must match")
        if len(self.dihedrals) != len(self.dihedral_types):
            raise Exception("len of dihedrals and dihedral_types must match")
        if len(self.impropers) != len(self.improper_types):
            raise Exception("len of impropers and improper_types must match")

        # these arrays must have at least as many atom_type_* entries as the num_atom_types. The
        # reason they do not have to match _exactly_ is because a subset of an Atoms object gets
        # _all_ the atom_type_* arrays and does not reindex the atom_types array, so the
        # num_atom_types may be fewer than the number of atom types in the atom_type_* arrays.
        if len(self.atom_type_labels) < self.num_atom_types:
            raise Exception("len of atom_type_labels (%d) must be >= num_atom_types (%d)" % (len(self.atom_type_labels), self.num_atom_types))
        if len(self.atom_type_elements) < self.num_atom_types:
            raise Exception("len of atom_type_elements (%d) must be >= num_atom_types (%d)" % (len(self.atom_type_elements), self.num_atom_types))
        if len(self.atom_type_masses) < self.num_atom_types:
            raise Exception("len of atom_type_masses (%d) must be >= num_atom_types (%d)" % (len(self.atom_type_masses), self.num_atom_types))

    @classmethod
    def load(cls, f, filetype=None, **kwargs):
        """Creates an Atoms object from either a path or a file object and filetype.

        Can load any of the supported types:

        - lammps data file: "lmpdat"
        - cif
        - cml

        Args:
            f (Str or Path or File): either a path to a file or an open File to load from
            filetype (Str): filetype ('lmpdat', 'cif', or 'cml') of passed f File object, or
                explicit filetype to override default filetype implied from file extension.
            kwargs: keyword args passed on to individual load functions.
        """

        fd = None
        path = None

        if isinstance(f, io.TextIOBase):
            fd = f
            if filetype is None:
                raise Exception("If a File object is passed, a filetype must be passed with it")
        else:
            # other cases are treated as either Pathlib path or strings
            path = f
            if filetype is None:
                _, filetype = os.path.splitext(path)
                filetype = filetype[1:]

        if filetype == "lmpdat":
            with use_or_open(fd, path) as fh:
                atoms = cls.load_lmpdat(fh, **kwargs)
                return atoms
        elif filetype == "cml":
            return cls.load_cml(fd or path, **kwargs)
        elif filetype == "cif":
            with use_or_open(fd, path) as fh:
                return cls.load_cif(fh, **kwargs)
        else:
            raise Exception("Unsupported filetype")

    def save(self, f, filetype=None, **kwargs):
        """Saves an Atoms object to either a path or a file object and filetype.

        Can save any of the supported types:

        - lammps data file: "lmpdat"
        - mol

        Args:
            f (Str or Path or File): either a path to a file or an open File to save to
            filetype (Str): filetype ('lmpdat', 'cif', or 'cml') of passed f File object, or
                explicit filetype to override default filetype implied from file extension.
            kwargs: keyword args passed on to individual save functions.
        """

        fd = None
        path = None

        if isinstance(f, io.TextIOBase):
            fd = f
            if filetype is None:
                raise Exception("If a File object is passed, a filetype must be passed with it")
        else:
            # other cases are treated as either Pathlib path or strings
            path = f
            if filetype is None:
                _, filetype = os.path.splitext(path)
                filetype = filetype[1:]

        if filetype == "lmpdat":
            with use_or_open(fd, path, mode='w') as fh:
                atoms = self.save_lmpdat(fh, **kwargs)
                return atoms
        elif filetype == "mol":
            with use_or_open(fd, path, mode='w') as fh:
                return self.save_mol(fh, **kwargs)
        else:
            raise Exception("Unsupported filetype")

    @classmethod
    def load_lmpdat(cls, f, atom_format="full", guess_atol=0.1):
        """Load Atoms object from lammps data file (.lmpdat) format.

        LAMMPS data files store only atom ids and masses, but do not store two other things we need:
        elements and atom type labels. These are the rules for inferring atom type labels and elements.

        In priority order, for elements, we:

        1. guess the elements using the masses by seeing if there is a periodic table element within
        0.1 g/mol of the mass. If any atom types doe not match to an existing periodic table
        element, this method fails.
        2. use the atom ids as the elements (and print a warning).

        In priority order, for atom type labels, we:

        1. use the comments after each line in the Masses section as the atom type. If any line is
        missing a comment, this method fails.
        2. use the elements, if we have them.
        3. use the atom ids (and print a warning).

        Args:
            f (File): File-like object to read from.
            atom_format (str): atom format of lammps data file. Currently supported atom formats are
                'full' and 'atomic'.
            guess_atol (float): absolute tolerance a read mass can differ from a periodic table mass
                and still be considered that element. Default: 0.1.

        Returns:
            Atoms: loaded Atoms object
        """
        def get_types_tups(arr):
            types = tups = []
            if len(arr) > 0:
                types = arr[:, 1] - 1
                tups = arr[:, 2:] - 1
            return types, tups

        masses = []
        atoms = []
        bonds = []
        angles = []
        dihedrals = []
        impropers = []

        pair_coeffs = []
        bond_coeffs = []
        angle_coeffs = []
        dihedral_coeffs = []
        improper_coeffs = []
        atom_type_labels = []
        atom_type_elements = []

        cellx = celly = cellz = 0.0

        sections_handled = ["Pair Coeffs", "Bond Coeffs", "Angle Coeffs", "Dihedral Coeffs",
                            "Improper Coeffs", "Atoms", "Bonds", "Angles", "Dihedrals", "Impropers",
                            "Masses"]
        current_section = None
        start_section = False

        for unprocessed_line in f:
            # handle comments
            comment_string = ""
            comment = None
            if "#" in unprocessed_line:
                line, comment = unprocessed_line.split('#')
                comment = comment.strip()
                comment_string ="   # " + comment
            else:
                line = unprocessed_line.split('#')[0]
            line = line.strip()
            if line in sections_handled:
                current_section = line
                start_section = True
                continue
            elif line == "":
                if not start_section:
                    # end of section or blank
                    current_section= None
                start_section = False
                continue
            tup = line.split()
            if current_section == "Masses":
                masses.append(tup[1])
                atom_type_labels.append(comment)
            elif current_section == "Pair Coeffs":
                pair_coeffs.append("%s%s" % (" ".join(tup[1:]), comment_string))
            elif current_section == "Bond Coeffs":
                bond_coeffs.append("%s%s" % (" ".join(tup[1:]), comment_string))
            elif current_section == "Angle Coeffs":
                angle_coeffs.append("%s%s" % ("  ".join(tup[1:]), comment_string))
            elif current_section == "Dihedral Coeffs":
                dihedral_coeffs.append("%s%s" % (" ".join(tup[1:]), comment_string))
            elif current_section == "Improper Coeffs":
                improper_coeffs.append("%s%s" % (" ".join(tup[1:]), comment_string))
            elif current_section == "Atoms":
                atoms.append(tup)
            elif current_section == "Bonds":
                bonds.append(tup)
            elif current_section == "Angles":
                angles.append(tup)
            elif current_section == "Dihedrals":
                dihedrals.append(tup)
            elif current_section == "Impropers":
                impropers.append(tup)
            elif current_section is None:
                if "xlo xhi" in line:
                    cellx = float(tup[1]) - float(tup[0])
                elif "ylo yhi" in line:
                    celly = float(tup[1]) - float(tup[0])
                elif "zlo zhi" in line:
                    cellz = float(tup[1]) - float(tup[0])

        cell = None
        if cellx > 0. and celly > 0. and cellz > 0.:
            cell = np.identity(3) * (cellx, celly, cellz)
        atom_type_masses = np.array(masses, dtype=float)
        atoms = np.array(atoms, dtype=float)
        bonds = np.array(bonds, dtype=int)
        angles = np.array(angles, dtype=int)
        dihedrals = np.array(dihedrals, dtype=int)
        impropers = np.array(impropers, dtype=int)

        # note: bond indices in lmpdat file are 1-indexed and we are 0-indexed which is why
        # the bond pairs get a -1
        if atom_format == "atomic":
            atom_types = np.array(atoms[:, 1] - 1, dtype=int)
            groups = np.zeros(len(atom_types))
            charges = np.zeros(len(atom_types))
            atom_tups = atoms[:, 2:5]
        elif atom_format == "full":
            groups = np.array(atoms[:, 1] - 1, dtype=int)
            atom_types = np.array(atoms[:, 2] - 1, dtype=int)
            charges = np.array(atoms[:, 3], dtype=float)
            atom_tups = atoms[:, 4:7]

        # guess the atom elements; if this fails, use the atoms ids as the elements
        try:
            atom_type_elements = guess_elements_from_masses(atom_type_masses, max_delta=guess_atol)
        except Exception:
            print("Warning: using type ids for elements since some masses do not correspond to periodic table elements within the set tolerance.")
            atom_type_elements = [str(i + 1) for i in range(len(masses))]

        # infer the atom type labels
        if (atom_type_labels.count(None) == 0):
            # then loading atom types from the labels worked
            pass
        else:
            print("Warning: using elements for atom type labels since there is not an atom type label comment for every atom type in the Masses section.")
            atom_type_labels = atom_type_elements.copy()


        bond_types, bond_tups = get_types_tups(bonds)
        angle_types, angle_tups = get_types_tups(angles)
        dihedral_types, dihedral_tups = get_types_tups(dihedrals)
        improper_types, improper_tups = get_types_tups(impropers)

        return cls(atom_types=atom_types, positions=atom_tups, charges=charges,
                   atom_type_masses=atom_type_masses, atom_type_elements=atom_type_elements,
                   bond_types=bond_types, bonds=bond_tups,
                   angle_types=angle_types, angles=angle_tups,
                   dihedral_types=dihedral_types, dihedrals=dihedral_tups,
                   improper_types=improper_types, impropers=improper_tups,
                   pair_coeffs=pair_coeffs, bond_type_coeffs=bond_coeffs,
                   angle_type_coeffs=angle_coeffs, dihedral_type_coeffs=dihedral_coeffs,
                   improper_type_coeffs=improper_coeffs, atom_type_labels=atom_type_labels,
                   groups=groups, cell=cell)

    def save_lmpdat(self, f, atom_format="full", file_comment=""):
        """Saves a lmpdat file

        Args:
            f (File): an open file to write to
            atom_format (str): LAMMPS atom format. Supports only 'atomic' and 'full'.
            file_comment (str): written in first line of output file.
        """

        f.write("%s (written by mofun)\n\n" % file_comment)

        f.write('%d atoms\n' % len(self.atom_types))
        f.write('%d bonds\n' % len(self.bond_types))
        f.write('%d angles\n' % len(self.angle_types))
        f.write('%d dihedrals\n' % len(self.dihedral_types))
        f.write('%d impropers\n' % len(self.improper_types))
        f.write("\n")

        if self.num_atom_types > 0:
            f.write('%d atom types\n' % self.num_atom_types)
        if self.num_bond_types > 0:
            f.write('%d bond types\n' % self.num_bond_types)
        if self.num_angle_types > 0:
            f.write('%d angle types\n' % self.num_angle_types)
        if self.num_dihedral_types > 0:
            f.write('%d dihedral types\n' % self.num_dihedral_types)
        if self.num_improper_types > 0:
            f.write('%d improper types\n' % self.num_improper_types)

        # TODO: support triclinic
        if self.cell is not None and self.cell.shape == (3,3):
            xlohi, ylohi, zlohi = zip([0,0,0], np.diag(self.cell))
            f.write(" %10.6f %10.6f xlo xhi\n" % xlohi)
            f.write(" %10.6f %10.6f ylo yhi\n" % ylohi)
            f.write(" %10.6f %10.6f zlo zhi\n" % zlohi)

        f.write("\nMasses\n\n")
        for i, m in enumerate(self.atom_type_masses):
            f.write(" %d %10.6f   # %s\n" % (i + 1, m, self.label_atoms(i)))

        if len(self.pair_coeffs) > 0:
            f.write('\nPair Coeffs\n\n')
            for i, coeffs in enumerate(self.pair_coeffs):
                f.write(' %d %s\n' % (i + 1, coeffs))

        if len(self.bond_type_coeffs) > 0:
            f.write('\nBond Coeffs\n\n')
            for i, coeffs in enumerate(self.bond_type_coeffs):
                f.write(' %d %s\n' % (i + 1, coeffs))

        if len(self.angle_type_coeffs) > 0:
            f.write('\nAngle Coeffs\n\n')
            for i, coeffs in enumerate(self.angle_type_coeffs):
                f.write(' %d %s\n' % (i + 1, coeffs))

        if len(self.dihedral_type_coeffs) > 0:
            f.write('\nDihedral Coeffs\n\n')
            for i, coeffs in enumerate(self.dihedral_type_coeffs):
                f.write(' %d %s\n' % (i + 1, coeffs))

        if len(self.improper_type_coeffs) > 0:
            f.write('\nImproper Coeffs\n\n')
            for i, coeffs in enumerate(self.improper_type_coeffs):
                f.write(' %d %s\n' % (i + 1, coeffs))

        f.write("\nAtoms\n\n")
        if atom_format == "atomic":
            for i, (x, y, z) in enumerate(self.positions):
                f.write(" %d %d %10.6f %10.6f %10.6f   # %s\n" % (i + 1, self.atom_types[i] + 1, x, y, z, self.label_atoms(self.atom_types[i])))
        elif atom_format == "full":
            for i, (x, y, z) in enumerate(self.positions):
                f.write(" %d %d %d %10.6f %10.6f %10.6f %10.6f   # %s\n" %
                    (i + 1, self.groups[i] + 1, self.atom_types[i] + 1, self.charges[i], x, y, z, self.label_atoms(self.atom_types[i])))

        if len(self.bonds) > 0:
            f.write("\nBonds\n\n")
            for i, tup in enumerate(self.bonds):
                f.write(" %d %d %d %d   # %s\n" % (i + 1, self.bond_types[i] + 1, *(np.array(tup) + 1), self.label_atoms(tup, atom_indices=True)))

        if len(self.angles) > 0:
            f.write("\nAngles\n\n")
            for i, tup in enumerate(self.angles):
                f.write(" %d %d %d %d %d   # %s\n" % (i + 1, self.angle_types[i] + 1, *(np.array(tup) + 1), self.label_atoms(tup, atom_indices=True)))

        if len(self.dihedrals) > 0:
            f.write("\nDihedrals\n\n")
            for i, tup in enumerate(self.dihedrals):
                f.write(" %d %d %d %d %d %d   # %s\n" % (i + 1, self.dihedral_types[i] + 1, *(np.array(tup) + 1), self.label_atoms(tup, atom_indices=True)))

        if len(self.impropers) > 0:
            f.write("\nImpropers\n\n")
            for i, tup in enumerate(self.impropers):
                f.write(" %d %d %d %d %d %d   # %s\n" % (i + 1, self.improper_types[i] + 1, *(np.array(tup) + 1), self.label_atoms(tup, atom_indices=True)))


    def save_mol(self, f, file_comment=""):
        """Writes .mol file for structural information."""

        f.write(" Molecule_name: %s\n" % file_comment)
        f.write("\n")
        f.write("  Coord_Info: Listed Cartesian None\n")
        f.write("        %d\n" % len(self))

        for i, (x, y, z) in enumerate(self.positions):
            f.write("%6d %10.4f %10.4f %10.4f  %5s %10.8f  0  0\n" % (i + 1, x, y, z,
                self.elements[i], self.charges[i]))

        if self.cell is not None:
            f.write("\n\n\n")
            f.write("  Fundcell_Info: Listed\n")
            f.write("        %10.4f       %10.4f       %10.4f\n" % tuple(np.diag(self.cell)))
            f.write("           90.0000          90.0000          90.0000\n")
            f.write("           0.00000          0.00000          0.00000\n")
            f.write("        %10.4f       %10.4f       %10.4f\n" % tuple(np.diag(self.cell)))


    @classmethod
    def load_cif(cls, f):
        """Loads a CIF file, including bonding information.

        Args:
            f (File): File-like object to read from.

        Returns:
            Atoms: loaded Atoms object
        """

        def has_all_tags(block, tags):
            return np.array([block.has_key(tag) for tag in tags]).all()

        # PyCifRw supports file descriptors and path strings, but doesn't not support PathLib paths.
        if isinstance(f, pathlib.PurePath):
            f = str(f)

        cf = read_cif(f)
        block = cf[cf.get_roots()[0][0]]

        cart_coord_tags = ["_atom_site_Cartn_x", "_atom_site_Cartn_y", "_atom_site_Cartn_z", "_atom_site_label"]
        fract_coord_tags = ["_atom_site_fract_x", "_atom_site_fract_y", "_atom_site_fract_z", "_atom_site_label"]
        use_fract_coords = False
        if has_all_tags(block, cart_coord_tags):
            coords = [block[lbl] for lbl in cart_coord_tags]
        elif has_all_tags(block, fract_coord_tags):
            use_fract_coords = True
            coords = [block[lbl] for lbl in fract_coord_tags]
        else:
            raise("no fractional or cartesian coords in CIF file")

        x = [float(c) for c in coords[0]]
        y = [float(c) for c in coords[1]]
        z = [float(c) for c in coords[2]]

        atom_name = coords[3]
        positions = np.array([x,y,z], dtype=float).T

        atom_types = block['_atom_site_type_symbol']

        charges = []
        if block.has_key('_atom_site_charge'):
            charges = block['_atom_site_charge']

        bonds = []
        bond_tags = ["_geom_bond_atom_site_label_1", "_geom_bond_atom_site_label_2"]
        if has_all_tags(block, bond_tags):
            cifbonds = zip(*[block[lbl] for lbl in bond_tags])
            bonds = [(atom_name.index(a), atom_name.index(b)) for (a,b) in cifbonds]
            print("WARNING: cif read doesn't handle bond types at present; bonding info is discarded. Use LAMMPS data file format if you need bonds")
            bonds = []

        cell=None
        cell_tags = ['_cell_length_a', '_cell_length_b', '_cell_length_c', '_cell_angle_alpha', '_cell_angle_beta', '_cell_angle_gamma']
        if has_all_tags(block, cell_tags):
            a, b, c, alpha, beta, gamma = [float(block[tag]) for tag in cell_tags]
            # TODO: Fix for triclinic
            if alpha != 90. or beta != 90 or gamma != 90.:
                raise Exception("No support for non orthorhombic UCs at the moment!")

            cell=np.identity(3) * (a, b, c)
            if use_fract_coords:
                positions *= (a,b,c)

        return cls(elements=atom_types, positions=positions, cell=cell, charges=charges)

    @classmethod
    def load_cml(cls, f, verbose=False):
        """Loads a CML file, including bonding information.

        Args:
            f (Path or File): Path or File-like object to read from.

        Returns:
            Atoms: loaded Atoms object
        """

        tree = ET.parse(f)
        root = tree.getroot()

        atom_dicts = [a.attrib for a in root.findall('.//atom')]
        atom_tuples = [(a['id'], a['elementType'],
                       float(a['x3']), float(a['y3']), float(a['z3'])) for a in atom_dicts]
        ids, elements, x, y, z = zip(*atom_tuples)
        id_to_idx = {id:i for i, id in enumerate(ids)}
        positions = np.array([x,y,z]).T

        bond_dicts = [a.attrib for a in root.findall('.//bond')]
        bond_tuples = [(a['atomRefs2'].split(), float(a['order'])) for a in bond_dicts]
        bonds_by_ids, bond_orders = zip(*bond_tuples)
        bonds = [(id_to_idx[b1], id_to_idx[b2]) for (b1,b2) in bonds_by_ids]
        bond_types = [0 for b in bonds]
        if verbose:
            print("Found %d atoms: %s" % (len(elements), elements))
            print("Found %d atom positions: %s" % (len(positions), positions))
            print("Found %d bonds: %s" % (len(bonds), bonds))
            print("Found %d bond_types: %s" % (len(bond_types), bond_types))

        return cls(elements=elements, positions=positions, bonds=bonds, bond_types=bond_types)

    @classmethod
    def from_ase_atoms(cls, atoms):
        """Create an Atoms object from an ASE Atoms object.

        Only supports importing the atom positions, elements, and the unit cell.

        Args:
            atoms (ASE Atoms object): atoms to load from

        Returns:
            Atoms: atoms loaded from an ASE Atoms object.
        """
        return cls(elements=atoms.symbols, positions=atoms.positions, cell=atoms.cell)


    def label_atoms(self, atoms, atom_indices=False):
        if not hasattr(atoms, '__iter__'):
            atoms = [atoms]

        if atom_indices:
            atoms = [self.atom_types[i] for i in atoms]

        return " ".join([self.atom_type_labels[x] for x in atoms])

    @property
    def elements(self):
        return [self.atom_type_elements[i] for i in self.atom_types]

    @property
    def symbols(self):
        return self.elements

    def copy(self):
        return copy.deepcopy(self)

    def translate(self, delta):
        if self.positions is not None and len(self) > 0:
            self.positions += delta

    def __len__(self):
        return len(self.positions)

    @property
    def num_atom_types(self):
        if len(self.atom_types) == 0:
            return 0
        return len(self.atom_type_elements)

    @property
    def num_bond_types(self):
        if len(self.bond_types) == 0:
            return 0
        return len(self.bond_type_coeffs) or max(self.bond_types) + 1

    @property
    def num_angle_types(self):
        if len(self.angle_types) == 0:
            return 0
        return len(self.angle_type_coeffs) or max(self.angle_types) + 1

    @property
    def num_dihedral_types(self):
        if len(self.dihedral_types) == 0:
            return 0
        return len(self.dihedral_type_coeffs) or max(self.dihedral_types) + 1

    @property
    def num_improper_types(self):
        if len(self.improper_types) == 0:
            return 0
        return len(self.improper_type_coeffs) or max(self.improper_types) + 1

    def extend_types(self, other):
        offsets = (self.num_atom_types, self.num_bond_types,
                   self.num_angle_types, self.num_dihedral_types, self.num_improper_types)

        self.atom_type_elements = np.append(self.atom_type_elements, other.atom_type_elements)
        self.atom_type_masses = np.append(self.atom_type_masses, other.atom_type_masses)
        self.atom_type_labels = np.append(self.atom_type_labels, other.atom_type_labels)
        self.pair_coeffs = np.append(self.pair_coeffs, other.pair_coeffs)

        self.bond_type_coeffs = np.append(self.bond_type_coeffs, other.bond_type_coeffs)
        self.angle_type_coeffs = np.append(self.angle_type_coeffs, other.angle_type_coeffs)
        self.dihedral_type_coeffs = np.append(self.dihedral_type_coeffs, other.dihedral_type_coeffs)
        self.improper_type_coeffs = np.append(self.improper_type_coeffs, other.improper_type_coeffs)

        return offsets


    def extend(self, other, offsets=None, structure_index_map={}, verbose=False):
        """Adds other Atoms object's arrays to its own.

        The default behavior is for all the types and params from other structure to be appended to
        this structure.

        Alternatively, an offsets tuple may be passed with the results of calling extend_types(). No
        new types will be added, but the newly added atoms, bonds, etc will refer to types by their
        value in the other Atoms object plus the offset. Use this when you are adding the same set
        of atoms multiple times, or if your other atoms already share the same type ids as this
        object. For the later case, the tuple (0,0,0,0) may be passed in.

        Args:
            other (Atoms): atoms to add to self
            offsets: an offsets tuple with the results of calling extend_types().
            structure_index_map: dictionary where key is an index in other and value is an index in
                self, where entries only exist if the position and element of the entries are
                identical and can be considered to be the same atom.
            verbose (bool): print debugging info.
        """

        atom_idx_offset = len(self.positions)
        if offsets is None:
            if verbose:
                print("auto offset: extending types")
            offsets = self.extend_types(other)

        # update atom types for atoms that are already part of self Atoms object
        for other_index, self_index in structure_index_map.items():
            self.atom_types[self_index] = other.atom_types[other_index] + offsets[0]

        # add atoms that are not part of self Atoms object
        atoms_to_add = [i for i in range(len(other)) if i not in structure_index_map.keys()]
        self.positions = np.append(self.positions, other.positions[atoms_to_add], axis=0)
        self.atom_types = np.append(self.atom_types, other.atom_types[atoms_to_add] + offsets[0])
        self.charges = np.append(self.charges, other.charges[atoms_to_add])
        self.groups = np.append(self.groups, other.groups[atoms_to_add])

        # update structure index map
        structure_index_map2 = {a:i + atom_idx_offset for i,a in enumerate(atoms_to_add)}
        structure_index_map2.update(structure_index_map)
        convert2structureindex = np.vectorize(structure_index_map2.get)

        def find_existing_topo(topo, new_topo):
            """ find existing topo tuples between the same atoms as a new topo set.

            Used to allow an override of an existing force field term by finding old terms between
            the same atoms to delete"""
            if len(topo) == 0:
                return []
            return list(np.nonzero(cdist(topo, new_topo, 'cityblock') == 0)[0])

        if len(other.bonds) > 0:
            new_bonds = convert2structureindex(other.bonds)
            existing_bond_indices = find_existing_topo(self.bonds, new_bonds)
            self.bonds = np.append(self.bonds, new_bonds).reshape((-1,2))
            self.bond_types = np.append(self.bond_types, other.bond_types + offsets[1])
            self.bonds = np.delete(self.bonds, existing_bond_indices, axis=0)
            self.bond_types = np.delete(self.bond_types, existing_bond_indices)

        if len(other.angles) > 0:
            new_angles = convert2structureindex(other.angles)
            existing_angle_indices = find_existing_topo(self.angles, new_angles)
            self.angles = np.append(self.angles, new_angles).reshape((-1,3))
            self.angle_types = np.append(self.angle_types, other.angle_types + offsets[2])
            self.angles = np.delete(self.angles, existing_angle_indices, axis=0)
            self.angle_types = np.delete(self.angle_types, existing_angle_indices)

        if len(other.dihedrals) > 0:
            new_dihedrals = convert2structureindex(other.dihedrals)
            existing_dihedral_indices = find_existing_topo(self.dihedrals, new_dihedrals)
            self.dihedrals = np.append(self.dihedrals, new_dihedrals).reshape((-1,4))
            self.dihedral_types = np.append(self.dihedral_types, other.dihedral_types + offsets[3])
            self.dihedrals = np.delete(self.dihedrals, existing_dihedral_indices, axis=0)
            self.dihedral_types = np.delete(self.dihedral_types, existing_dihedral_indices)

        if len(other.impropers) > 0:
            new_impropers = convert2structureindex(other.impropers)
            existing_improper_indices = find_existing_topo(self.impropers, new_impropers)
            self.impropers = np.append(self.impropers, new_impropers).reshape((-1,4))
            self.improper_types = np.append(self.improper_types, other.improper_types + offsets[4])
            self.impropers = np.delete(self.impropers, existing_improper_indices, axis=0)
            self.improper_types = np.delete(self.improper_types, existing_improper_indices)

        self.assert_arrays_are_consistent_sizes()

    def replicate(self, repldims=(1,1,1)):
        """Replicate atoms object across xyz dimensions

        Warnings:
        * does not magically handle any bonds that may cross periodic boundary conditions!

        Args:
            repldims (Tuple(int, int, int)): number of times to replicate in each dimension

        Returns:
            Atoms: replicated atoms.
        """
        if self.cell is None:
            raise Exception("Can't replicate if no unit cell has been defined")

        repl_atoms = self.copy()
        ucmults = np.array(np.meshgrid(*[range(r) for r in repldims])).T.reshape(-1, 3)
        ucmults = ucmults[np.any(ucmults != 0, axis=1)] # remove [0,0,0] since in copy
        for ucmult in ucmults:
            transatoms = self.copy()
            transatoms.translate(np.matmul(transatoms.cell.T, ucmult))
            repl_atoms.extend(transatoms, offsets=(0,0,0,0))

        repl_atoms.cell = self.cell * repldims
        return repl_atoms

    def _delete_and_reindex_atom_index_array(self, arr, sorted_deleted_indices, secondary_arr=None):
        updated_arr = arr.copy()
        arr_idx_to_delete = []
        for i, atom_idx_tuple in enumerate(arr):
            if np.any([a in sorted_deleted_indices for a in atom_idx_tuple]):
                arr_idx_to_delete.append(i)
        updated_arr = np.delete(updated_arr, arr_idx_to_delete, axis=0)

        # reindex
        for i in sorted_deleted_indices:
            np.subtract(updated_arr, 1, out=updated_arr, where=updated_arr>i)

        if secondary_arr is not None:
            # remove same indices from secondary array; this is used for types arrays
            secondary_arr = np.delete(secondary_arr, arr_idx_to_delete, axis=0)
            return (updated_arr, secondary_arr)
        else:
            return updated_arr

    def __delitem__(self, indices):
        self.positions = np.delete(self.positions, indices, axis=0)
        self.atom_types = np.delete(self.atom_types, indices, axis=0)
        self.charges = np.delete(self.charges, indices, axis=0)
        self.groups = np.delete(self.groups, indices, axis=0)

        sorted_indices = sorted(indices, reverse=True)
        if len(self.bonds) > 0:
            self.bonds, self.bond_types = self._delete_and_reindex_atom_index_array(self.bonds, sorted_indices, self.bond_types)
        if len(self.angles) > 0:
            self.angles, self.angle_types = self._delete_and_reindex_atom_index_array(self.angles, sorted_indices, self.angle_types)
        if len(self.dihedrals) > 0:
            self.dihedrals, self.dihedral_types = self._delete_and_reindex_atom_index_array(self.dihedrals, sorted_indices, self.dihedral_types)
        if len(self.impropers) > 0:
            self.impropers, self.improper_types = self._delete_and_reindex_atom_index_array(self.impropers, sorted_indices, self.improper_types)

        self.assert_arrays_are_consistent_sizes()

    def pop(self, pos=-1):
        del(self, pos)

    def __getitem__(self, i):
        idx = np.array(i, ndmin=1)
        return Atoms(positions=np.take(self.positions, idx, axis=0),
                     atom_types=np.take(self.atom_types, idx, axis=0),
                     charges=np.take(self.charges, idx, axis=0),
                     atom_type_masses=self.atom_type_masses,
                     atom_type_elements=self.atom_type_elements,
                     groups=np.take(self.groups, idx, axis=0),
                     cell=self.cell)

    def to_ase(self):
        """Convert to ASE atoms object.

        Only supports export of the positions and elements.
        """
        kwargs = dict(positions=self.positions)
        if self.cell is not None:
            kwargs['cell'] = self.cell
            kwargs['pbc'] = True
        return ase.Atoms(self.elements, **kwargs)

def find_unchanged_atom_pairs(orig_structure, final_structure, max_delta=1e-5):
    """Returns array of tuple pairs, where each pair contains the indices in the original and the final
    structure that match.

    Does not work across PBCs."""
    match_pairs = []
    for i, p1 in enumerate(orig_structure.positions):
        for j, p2 in enumerate(final_structure.positions):
            if norm(np.array(p2) - p1) < max_delta and orig_structure.elements[i] == final_structure.elements[j]:
                match_pairs.append((i,j))
                break
    return match_pairs




import pytest

pytest.register_assert_rewrite("tests.fixtures")

def pytest_addoption(parser):
    parser.addoption(
        "--runslow", action="store_true", default=False, help="run slow tests"
    )
    parser.addoption(
        "--runveryslow", action="store_true", default=False, help="run slow tests"
    )

def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark test as slow to run")
    config.addinivalue_line("markers", "veryslow: mark test as VERY slow to run")

def pytest_collection_modifyitems(config, items):
    runveryslow = config.getoption("--runveryslow")
    runslow = config.getoption("--runslow") or runveryslow

    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    skip_very_slow = pytest.mark.skip(reason="need --runveryslow option to run")

    for item in items:
        if "slow" in item.keywords and not runslow:
            item.add_marker(skip_slow)
        if "veryslow" in item.keywords and not runveryslow:
            item.add_marker(skip_very_slow)



"""
Take CIF file, optionally update with charges from chargefile, replicate to meet MIC, assign pair
potentials and output to a working LAMMPS Data file.
"""

import pathlib

import click
import numpy as np
from mofun import Atoms, replace_pattern_in_structure, find_pattern_in_structure
from mofun.rough_uff import pair_coeffs
from mofun.uff4mof import uff_key_starts_with

@click.command()
@click.argument('inputpath', type=click.Path(path_type=pathlib.Path))
@click.argument('outputpath', type=click.Path(path_type=pathlib.Path))
@click.option('-f', '--find', 'find_path', type=click.Path(path_type=pathlib.Path))
@click.option('-r', '--replace', 'replace_path', type=click.Path(path_type=pathlib.Path))
@click.option('-p', '--replace-fraction', type=float, default=1.0)
@click.option('--atol', type=float, default=5e-2, help="absolute tolerance in Angstroms for atom posistions to be considered matching")
@click.option('-ap1', '--axisp1-idx', type=int, default=0, help="index of first point on primary rotation axis")
@click.option('-ap2', '--axisp2-idx', type=int, default=-1, help="index of second point on primary rotation axis")
@click.option('-op', '--opoint-idx', type=int, default=None, help="index of point that makes up secondary rotation axis (between this point and the primary rotation axis)")
@click.option('--dumppath', type=click.Path(path_type=pathlib.Path))
@click.option('-q', '--chargefile', type=click.File('r'))
@click.option('--replicate', nargs=3, type=int, help="replicate structure across x, y, and z dimensions")
@click.option('--mic', type=float, help="enforce minimum image convention using a cutoff of mic")
@click.option('--framework-element', type=str, help="convert all atoms that are in group 0, the framework group to a specific atom type to make vizualizing the structure easier")
@click.option('--pp', is_flag=True, default=False, help="Assign UFF pair potentials to atoms (sufficient for fixed force-field calculations)")
def mofun_cli(inputpath, outputpath,
        find_path=None, replace_path=None, atol=5e-2, replace_fraction=1.0, axisp1_idx=0, axisp2_idx=-1, opoint_idx=None,
        dumppath=None, chargefile=None, replicate=None, mic=None, framework_element=None, pp=False):
    atoms = Atoms.load(inputpath)

    # upate positions from lammps dump file
    if dumppath is not None:
        # update positions in original atoms file with new positions
        dumpatoms = ase.io.read(dumppath, format="lammps-dump-text")
        assert len(dumpatoms.positions) == len(atoms.positions)
        atoms.positions = dumpatoms.positions

    # update charges
    if chargefile is not None:
        charges = np.array([float(line.strip()) for line in chargefile if line.strip() != ''])
        assert len(charges) == len(atoms.positions)
        atoms.charges = charges

    if replicate is not None:
        atoms = atoms.replicate(replicate)

    # replicate to meet minimum image convention, if necessary
    if mic is not None:
        repls = np.array(np.ceil(2*mic / np.diag(atoms.cell)), dtype=int)
        atoms = atoms.replicate(repls)

    if pp:
        assign_pair_params_to_structure(atoms)

    if replace_path is not None and find_path is None:
        print("Cannot perform a replace operation without a find operation")
    elif find_path is not None:
        search_pattern = Atoms.load(find_path)
        if replace_path is not None:
            replace_pattern = Atoms.load(replace_path)
            atoms = replace_pattern_in_structure(atoms, search_pattern, replace_pattern, atol=atol,
                axisp1_idx=axisp1_idx, axisp2_idx=axisp2_idx, opoint_idx=opoint_idx, replace_fraction=replace_fraction)
        else:
            results = find_pattern_in_structure(atoms, search_pattern, atol=atol)
            print("Found %d instances of the search_pattern in the structure" % len(results))
            print(results)

    # set framework elements to specified element-only works on ASE exports
    if framework_element is not None:
        atoms.symbols[atoms.atom_groups == 0] = framework_element

    if outputpath.suffix in ['.lmpdat', '.mol']:
        atoms.save(outputpath)
    else:
        print("INFO: Trying output using ASE")
        aseatoms = atoms.to_ase()
        if framework_element is not None:
            aseatoms.symbols[atoms.atom_groups == 0] = framework_element

        aseatoms.set_pbc(True)
        aseatoms.write(outputpath)

def assign_pair_params_to_structure(structure):
    # NOTE: in UFF, pair params should always be the same for atoms of the same element, regardless of type
    # DUPLICATE in functionalize_linkers.py: should be refactored
    uff_keys = [uff_key_starts_with(el.ljust(2, "_"))[0] for el in structure.atom_type_elements]
    structure.pair_coeffs = ['%10.6f %10.6f # %s' % (*pair_coeffs(k), k) for k in uff_keys]
    structure.atom_type_labels = uff_keys



import math
import random

import numpy as np
from scipy.spatial import distance
from scipy.spatial.transform import Rotation as R

from mofun.atoms import find_unchanged_atom_pairs
from mofun.helpers import atoms_of_type, atoms_by_type_dict, position_index_farthest_from_axis, \
                          quaternion_from_two_vectors, quaternion_from_two_vectors_around_axis, \
                          remove_duplicates, assert_positions_are_unchanged, suppress_warnings, group_duplicates, positions_are_unchanged

def uc_neighbor_offsets(uc_vectors):
    multipliers = np.array(np.meshgrid([-1, 0, 1],[-1, 0, 1],[-1, 0, 1])).T.reshape(-1, 1, 3)
    return (uc_vectors * multipliers).sum(axis=1)

def get_types_ss_map_limited_near_uc(structure, length):
    """
    structure:
    length: the length of the longest dimension of the search pattern

    creates master lists of indices, types and positions, for all atoms in the structure and all
    atoms across the PBCs. Limits atoms across PBCs to those that are within a distance of the
    boundary that is less than the length of the search pattern (i.e. atoms further away from the
    boundary than this will never match the search pattern).
    """

    cell = structure.cell
    uc_offsets = uc_neighbor_offsets(cell)
    # move (0., 0., 0.) to be at the 0 index
    uc_offsets[np.where(np.all(uc_offsets == (0,0,0), axis=1))[0][0]] = uc_offsets[0]
    uc_offsets[0] = (0.0, 0.0, 0.0)

    s_positions = [structure.positions + uc_offset for uc_offset in uc_offsets]
    s_positions = np.array([x for y in s_positions for x in y])

    s_types = list(structure.elements) * len(uc_offsets)

    index_mapper = []
    s_pos_view = []
    s_types_view = []

    is_triclinic = not (np.diag(cell) * np.identity(3) == cell).all()
    if is_triclinic:
        # search within triclinic space + buffer by looking at three planes that go through origin

        # normal vectors for planes: xy, xz, yz
        nvs = np.array([np.cross(cell[0], cell[1]), np.cross(cell[0], cell[2]), np.cross(cell[1], cell[2])])
        nvnorms = np.linalg.norm(nvs, axis=1)
        planedists = np.abs(np.array([np.dot(cell[2], nvs[0]) / nvnorms[0],
                                      np.dot(cell[1], nvs[1]) / nvnorms[1],
                                      np.dot(cell[0], nvs[2]) / nvnorms[2]]))
        # calculate distance to center point; from a plane boundary to inside the unit cell
        # (inwards) should have a negative distance. If not, we will multiply the distances below
        # by -1 to account for this.
        centerpos = cell.sum(axis=0) / 2
        centerdist = np.dot(nvs, centerpos)
        nmults = -centerdist / np.abs(centerdist)

        for i, pos in enumerate(s_positions):
            if ((-planedists[0] - length <= nmults[0] * np.dot(nvs[0], pos) / nvnorms[0] <= length) and
                (-planedists[1] - length <= nmults[1] * np.dot(nvs[1], pos) / nvnorms[1] <= length) and
                (-planedists[2] - length <= nmults[2] * np.dot(nvs[2], pos) / nvnorms[2] <= length)):

                index_mapper.append(i)
                s_pos_view.append(pos)
                s_types_view.append(s_types[i])

    else: # orthorhombic
        cell = list(np.diag(cell))

        for i, pos in enumerate(s_positions):
            if (pos[0] >= -length and pos[0] < length + cell[0] and
                pos[1] >= -length and pos[1] < length + cell[1] and
                pos[2] >= -length and pos[2] < length + cell[2]):

                index_mapper.append(i)
                s_pos_view.append(pos)
                s_types_view.append(s_types[i])

    return s_types_view, index_mapper, s_pos_view, s_positions

@suppress_warnings
def find_pattern_in_structure(structure, pattern, axisp1_idx=0, axisp2_idx=-1, opoint_idx=None,
        return_positions_and_quats=False, atol=5e-2, verbose=False):
    """Looks for instances of `pattern` in `structure`, where a match in the structure has the same number of atoms, the
    same elements and the same relative coordinates as in `pattern`.

    Returns a list of tuples, one tuple per match found in `structure` where each tuple has the size `len(pattern)` and
    contains the indices in the structure that matched the pattern. If `return_positions_and_quats=True` then two
    additional lists are returned containing positions for each matched index for each match, and the quaternions
    required to rotate the search pattern to the match pattern.

    Args:
        structure (Atoms): an Atoms object to search in.
        pattern (Atoms): an Atoms object to search for.
        axisp1_idx (float): index in search_pattern of first point defining the directional axis of the search_pattern. May help with performance for large problems.
        axisp2_idx (float): index in search_pattern of second point defining the directional axis of the search_pattern. May help with performance for large problems.
        opoint_idx (float): index in search_pattern of the orientation point that we will use to align the search pattern (the search pattern will be rotated so that the orientation point in the search pattern will have the same coordinates as the same point in the match pattern).
        return_positions_and_quats (bool): additionally returns the positions for each index and the quaternions that rotation the search pattern to the match pattern.
        atol (float): the absolute tolerance (how close an atom must be in the structure to the position in pattern to be consdired a match).
        verbose (bool): print debugging info.
    Returns:
        List [tuple(len(pattern))],  {List [tuple(len(pattern))], List [scipy.spatial.transform.Rotation]} : returns a
            list of tuples of size `len(pattern)` containing the indices in the structure that matched the pattern, one
            tuple per each match, and optionally (if return_positions_and_quats is True) a corresponding list of tuples
            with positions instead of indices, and a list of quaternions (scipy.spatial.transform.Rotation), one per
            each match.
    """

    if verbose:
        print("calculating point distances...")
    p_ss = distance.cdist(pattern.positions, pattern.positions, "sqeuclidean")
    pattern_length = p_ss.max() ** 0.5 + 2 * atol
    s_types_view, index_mapper, s_pos_view, s_positions = get_types_ss_map_limited_near_uc(structure, pattern_length)
    atoms_by_type = atoms_by_type_dict(s_types_view)

    # created sorted coords array for creating search subsets
    p = np.array(sorted([(*r, i) for i, r in enumerate(s_pos_view)]))

    # Search instances of first atom in a search pattern
    # 0,0,0 uc atoms are always indexed first from 0 to # atoms in structure.
    starting_atoms = [idx for idx in atoms_of_type(s_types_view[0: len(structure)], pattern.elements[0])]
    if verbose:
        print("round %d (%d) [%s]: " % (0, len(starting_atoms), pattern.elements[0]), starting_atoms)

    def get_nearby_atoms(p, s_pos_view, pattern_length, a):
        p1 = p[(p[:, 0] <= s_pos_view[a][0] + pattern_length) & (p[:, 0] >= s_pos_view[a][0] - pattern_length)]
        p2 = p1[(p1[:, 1] <= s_pos_view[a][1] + pattern_length) & (p1[:, 1] >= s_pos_view[a][1] - pattern_length)]
        return (p2[(p2[:, 2] <= s_pos_view[a][2] + pattern_length) & (p2[:, 2] >= s_pos_view[a][2] - pattern_length)])

    pattern_elements = pattern.elements
    all_match_index_tuples = []
    for a_idx, a in enumerate(starting_atoms):
        match_index_tuples = [[a]]

        nearby = get_nearby_atoms(p, s_pos_view, pattern_length, a)
        nearby_atom_indices = nearby[:,3].astype(np.int32)
        nearby_positions = nearby[:,0:3]

        idx2ssidx = {atom_idx:i for i, atom_idx in enumerate(nearby_atom_indices)}
        s_ss = distance.cdist(nearby_positions, nearby_positions, "sqeuclidean")

        # start for loop at one since we've searched for starting atoms (index == 0) above
        for i in range(1, len(pattern)):
            if len(match_index_tuples) == 0:
                break
            last_match_index_tuples = match_index_tuples
            match_index_tuples = []
            for match in last_match_index_tuples:
                for ss_idx, atom_idx in enumerate(nearby_atom_indices):
                    if s_types_view[atom_idx]==pattern_elements[i]:
                        found_match = True
                        # check all distances to this new proposed atom
                        for j in range(0, i):
                            if not math.isclose(p_ss[i,j]**0.5, s_ss[idx2ssidx[match[j]], ss_idx]**0.5, abs_tol=atol):
                                found_match = False
                                break

                        # anything that matches the distance to all prior pattern atoms is a good match so far
                        if found_match:
                            match_index_tuples.append(match + [atom_idx])
            if verbose:
                print("round %d (%d) [%s]: " % (i, len(match_index_tuples), pattern.elements[i]), match_index_tuples)
        if verbose:
            print("starting atom %d: found %d matches: %s" % (a_idx, len(match_index_tuples), match_index_tuples))
        all_match_index_tuples += match_index_tuples

    ## remove duplicates by calculating quaternions necessary to align atoms, applying it to the search pattern and if
    # there is symmetry or chirality, eliminating possible matches where the atoms rotate into different places than we
    # found them in the structure.
    pattern = pattern.copy()
    pattern.translate(-pattern.positions[axisp1_idx])
    search_axis = pattern.positions[axisp2_idx]
    if len(pattern) > 2 and opoint_idx is None:
        # note that we find an orientation point if there are nore than two atoms in the search pattern, but it is still
        # possible that all the atoms like on the search axis. In that, case a point on the axis will be returned and
        # there is an unnecessary final rotation to "align" that point.
        opoint_idx = position_index_farthest_from_axis(search_axis, pattern)

    grouped_tuples = group_duplicates(all_match_index_tuples, key=lambda m: tuple(sorted([index_mapper[i] % len(structure) for i in m])))
    grouped_tuples2 = []

    good_match_index_tuples = []
    good_match_quats = []
    for _, match_tuples in grouped_tuples.items():
        quats = []
        good_indices = []
        for i, match_tuple in enumerate(match_tuples):
            atom_positions = np.array([s_positions[index_mapper[m]] for m in match_tuple])
            q = R.identity()
            if len(atom_positions) > 1:
                # the first quaternion aligns the search pattern axis points with the axis points
                # found in the structure
                match_axis = atom_positions[axisp2_idx] - atom_positions[axisp1_idx]
                q = quaternion_from_two_vectors(search_axis, match_axis)

                if len(atom_positions) > 2:
                    # the second quaternion is a rotation around the found axis in the structure and
                    # aligns the orientation axis point to its placement in the structure.
                    match_orientation_point = atom_positions[opoint_idx] - atom_positions[axisp1_idx]
                    rotated_orientation_point = q.apply(pattern.positions[opoint_idx])
                    q = quaternion_from_two_vectors_around_axis(rotated_orientation_point, match_orientation_point, match_axis) * q

            quats.append(q)
            chk_pattern = pattern.copy()
            chk_pattern.positions = q.apply(chk_pattern.positions)
            chk_pattern.translate(atom_positions[axisp1_idx])
            if positions_are_unchanged(atom_positions, chk_pattern.positions, max_delta=atol, verbose=verbose):
                good_indices.append(i)

        if len(good_indices) > 1:
            # Likely it is because of symmetry if we found more than one good match. Randomly choose one.
            match_chosen = random.choice(good_indices)
            good_match_index_tuples.append(match_tuples[match_chosen])
            good_match_quats.append(quats[match_chosen])
        elif len(good_indices) == 1:
            # Found one good match; either we have no symmetry, or we have symmetry where only one numbering of the
            # atoms can be rotated into place
            good_match_index_tuples.append(match_tuples[good_indices[0]])
            good_match_quats.append(quats[good_indices[0]])
        else:
            print("""WARNING:
                Search pattern was matched, but there is no possible way to rotate the search patttern to meet the match pattern.
                This is likely due to finding a match of the opposite chirality.
                """)

    match_index_tuples_in_uc = [tuple([index_mapper[m] % len(structure) for m in match]) for match in good_match_index_tuples]
    if return_positions_and_quats:
        match_index_tuple_positions = np.array([[s_positions[index_mapper[m]] for m in match] for match in good_match_index_tuples])
        return match_index_tuples_in_uc, match_index_tuple_positions, np.array(good_match_quats)
    else:
        return match_index_tuples_in_uc

class AtomsShouldNotBeDeletedTwice(Exception):
    pass

@suppress_warnings
def replace_pattern_in_structure(
    structure, search_pattern, replace_pattern, replace_fraction=1.0, atol=5e-2,
    axisp1_idx=0, axisp2_idx=-1, opoint_idx=None,
    return_num_matches=False, replace_all=False, verbose=False,
    positions_check_max_delta=0.1,
    ignore_atoms_should_not_be_deleted_twice=False):
    """Replaces all instances of `pattern` in `structure` with the `replace_pattern`.

    Works across periodic boundary conditions.

    WARNING: the replace pattern _MUST_ be on the same coordinate system as the search_pattern. If
    there are atoms that remain the same between the search and replace patterns, they must have the
    exact same coordinates. If these were to be offset, or moved, then when the replacement pattern
    gets inserted into the structure, then the replacement will also be offset.

    Args:
        structure (Atoms): an Atoms object to search in.
        search_pattern (Atoms): an Atoms object to search for.
        replace_pattern (Atoms): an Atoms object to search for.
        replace_fraction (float): how many instances of the search_pattern found in the structure get replaced by the replace pattern.
        atol (float): absolute tolerance in Angstroms for atom posistions to be considered matching.
        axisp1_idx (float): index in search_pattern of first point defining the directional axis of the search_pattern.
        axisp2_idx (float): index in search_pattern of second point defining the directional axis of the search_pattern.
        opoint_idx (float): index in search_pattern of third point defining the orientational axis of the search_pattern.
        replace_all (bool): replaces all atoms even if positions and elements match exactly
        verbose (bool): print debugging info.
        ignore_atoms_should_not_be_deleted_twice (bool): don't raise an AtomsShouldNotBeDeletedTwice exception when
            two matches would delete the same atoms.
    Returns:
        Atoms: the structure after search_pattern is replaced by replace_pattern.
    """
    search_pattern = search_pattern.copy()
    replace_pattern = replace_pattern.copy()

    # translate both search and replace patterns so that first atom of search pattern is at the origin
    replace_pattern.translate(-search_pattern.positions[axisp1_idx])
    search_pattern.translate(-search_pattern.positions[axisp1_idx])

    match_indices, match_positions, quats = find_pattern_in_structure(structure, search_pattern, atol=atol,
        axisp1_idx=axisp1_idx, axisp2_idx=axisp2_idx, opoint_idx=opoint_idx, return_positions_and_quats=True,
        verbose=verbose)

    if replace_fraction < 1.0:
        replace_indices = random.sample(list(range(len(match_positions))), k=round(replace_fraction * len(match_positions)))
        match_indices = [match_indices[i] for i in replace_indices]
        match_positions = match_positions[replace_indices]
        quats = quats[replace_indices]

    if verbose: print("match_indices / positions: ", match_indices, match_positions)

    replace2search_pattern_map = {k:v for (k,v) in find_unchanged_atom_pairs(replace_pattern, search_pattern)}

    new_structure = structure.copy()
    to_delete = set()
    if len(replace_pattern) == 0:
        to_delete |= set([idx for match in match_indices for idx in match])
    else:
        offsets = new_structure.extend_types(replace_pattern)

        for m_i, atom_positions in enumerate(match_positions):
            q = quats[m_i]
            new_atoms = replace_pattern.copy()
            new_atoms.positions = q.apply(new_atoms.positions)
            new_atoms.translate(atom_positions[axisp1_idx])
            new_atoms.positions %= np.diag(new_structure.cell)

            if verbose:
                print("new atoms after translate:\n", new_atoms.positions)

            structure_index_map = {}
            if not replace_all:
                structure_index_map = {k: match_indices[m_i][v] for k,v in replace2search_pattern_map.items()}
                new_structure.extend(new_atoms, offsets=offsets, structure_index_map=structure_index_map)
            else:
                new_structure.extend(new_atoms, offsets=offsets)

            to_delete_linker = set(match_indices[m_i]) - set(structure_index_map.values())
            if (to_delete.isdisjoint(to_delete_linker) or ignore_atoms_should_not_be_deleted_twice):
                to_delete |= set(to_delete_linker)
            else:
                raise AtomsShouldNotBeDeletedTwice()

    del(new_structure[list(to_delete)])

    if return_num_matches:
        return new_structure, len(match_indices)
    else:
        return new_structure



import io
from io import StringIO
from pathlib import Path

from tests.fixtures import *
# import tests.test_atoms_load_save
from mofun import Atoms


###### load/save LMPDAT format
def check_uio66_arb_terms_arrays(atoms):
    assert len(atoms.atom_type_masses) == 4
    assert len(atoms.pair_coeffs) == 4
    assert len(atoms.bond_type_coeffs) == 1
    assert len(atoms.angle_type_coeffs) == 2
    assert len(atoms.dihedral_type_coeffs) == 3
    assert len(atoms.improper_type_coeffs) == 1
    assert atoms.positions.shape == (16,3)
    assert len(atoms.groups) == 16
    assert len(atoms.charges) == 16
    assert len(atoms.atom_types) == 16
    assert len(atoms.bonds) == 4
    assert len(atoms.angles) == 8
    assert len(atoms.dihedrals) == 14
    assert len(atoms.impropers) == 2

def test_atoms_load_lmpdat__has_arrays_of_right_size():
    with Path("tests/uio66/uio66-linker-arb-terms.lmpdat").open() as f:
        atoms = Atoms.load_lmpdat(f, atom_format="full")

    check_uio66_arb_terms_arrays(atoms)


def test_atoms_load_lmpdat__atomic_masses_are_guessed_correctly():
    with Path("tests/test_atoms_load_save/atomic-masses.lmpdat").open() as f:
        atoms = Atoms.load_lmpdat(f, atom_format="full")
    assert np.array_equal(atoms.atom_type_elements, ["C", "N"])

def test_atoms_load_lmpdat__if_masses_are_nonatomic_then_elements_are_type_ids():
    with Path("tests/test_atoms_load_save/nonatomic-masses.lmpdat").open() as f:
        atoms = Atoms.load_lmpdat(f, atom_format="full")
    assert np.array_equal(atoms.atom_type_masses, [12., 1000.])
    assert np.array_equal(atoms.atom_type_elements, ["1", "2"])

def test_atoms_load_lmpdat__atom_type_labels_are_loaded_from_comments():
    with Path("tests/test_atoms_load_save/atomic-masses.lmpdat").open() as f:
        atoms = Atoms.load_lmpdat(f, atom_format="full")
    assert np.array_equal(atoms.atom_type_masses, [12., 14.])
    assert np.array_equal(atoms.atom_type_labels, ["C1", "N1"])

def test_atoms_load_lmpdat__atom_type_labels_are_loaded_from_elements_if_no_comments():
    with Path("tests/test_atoms_load_save/atomic-masses-no-comments.lmpdat").open() as f:
        atoms = Atoms.load_lmpdat(f, atom_format="full")
    assert np.array_equal(atoms.atom_type_masses, [12., 14.])
    assert np.array_equal(atoms.atom_type_labels, ["C", "N"])

def test_atoms_load_lmpdat__if_masses_are_nonatomic_with_no_comments_then_elements_and_labels_are_type_ids():
    with Path("tests/test_atoms_load_save/nonatomic-masses-no-comments.lmpdat").open() as f:
        atoms = Atoms.load_lmpdat(f, atom_format="full")
    assert np.array_equal(atoms.atom_type_masses, [12., 1000.])
    assert np.array_equal(atoms.atom_type_elements, ["1", "2"])
    assert np.array_equal(atoms.atom_type_labels, ["1", "2"])

def test_atoms_save_lmpdat__outputs_file_identical_to_input_file():
    with Path("tests/uio66/uio66-linker-arb-terms.lmpdat").open() as f:
        sin = StringIO(f.read())
        uio66_linker_ld = Atoms.load_lmpdat(sin, atom_format="full")

    sout = io.StringIO("")
    uio66_linker_ld.save_lmpdat(sout, file_comment="uio66-linker-arb-terms.lmpdat")

    # output file code, in case we need to update the lmpdat file because of new format changes
    # with open("uio66-linker-arb-terms.lmpdat", "w") as f:
    #     sout.seek(0)
    #     f.write(sout.read())

    sout.seek(0)
    sin.seek(0)
    assert sout.read() == sin.read()

###### load CML format
def test_atoms_load_cml__w_path_loads_elements_bonds():
    atoms = Atoms.load_cml("tests/uio66/uio66-linker.cml")

    assert atoms.elements == ["C", "O", "O", "C", "C", "H", "H", "C", "C", "C", "C", "H", "H", "O", "C", "O"]
    assert (atoms.bonds == [[0, 1], [10, 11], [0, 2], [0, 3], [3, 10], [9, 10], [3, 4], [9, 12],
                           [8, 9], [4, 5], [4, 7], [7, 8], [8, 14], [6, 7], [13, 14], [14, 15]]).all()

def test_atoms_load_cml__w_file_loads_elements_bonds():
    with open("tests/uio66/uio66-linker.cml", 'r') as f:
        atoms = Atoms.load_cml(f)

    assert atoms.elements == ["C", "O", "O", "C", "C", "H", "H", "C", "C", "C", "C", "H", "H", "O", "C", "O"]
    assert (atoms.bonds == [[0, 1], [10, 11], [0, 2], [0, 3], [3, 10], [9, 10], [3, 4], [9, 12],
                           [8, 9], [4, 5], [4, 7], [7, 8], [8, 14], [6, 7], [13, 14], [14, 15]]).all()

###### load CIF format
def test_atoms_load_cif__loads_elements():
    with Path("tests/uio66/uio66-linker.cif").open() as fd:
        atoms = Atoms.load_cif(fd)

    assert atoms.elements == ["C", "O", "O", "C", "C", "H", "H", "C", "C", "C", "C", "H", "H", "O", "C", "O"]

###### load method
def test_atoms_load__loads_lmpdat_from_file_or_path():
    with Path("tests/uio66/uio66-linker.lmpdat") as path:
        atoms = Atoms.load(path, atom_format="atomic")
        assert len(atoms) == 16

    with Path("tests/uio66/uio66-linker.lmpdat").open() as fd:
        atoms = Atoms.load(fd, filetype="lmpdat", atom_format="atomic")
        assert len(atoms) == 16

def test_atoms_load__loads_cml_from_file_or_path():
    with Path("tests/uio66/uio66-linker.cml") as path:
        atoms = Atoms.load(path)
        assert len(atoms) == 16

    with Path("tests/uio66/uio66-linker.cml").open() as fd:
        atoms = Atoms.load(fd, filetype="cml")
        assert len(atoms) == 16

def test_atoms_load__loads_cif_from_file_or_path():
    with Path("tests/uio66/uio66-linker.cif") as path:
        atoms = Atoms.load(path)
        assert len(atoms) == 16

    with Path("tests/uio66/uio66-linker.cif").open() as fd:
        atoms = Atoms.load(fd, filetype="cif")
        assert len(atoms) == 16



import warnings

import numpy as np
from pytest import approx

from mofun import quaternion_from_two_vectors, quaternion_from_two_vectors_around_axis


def test_quaternion_from_two_axes__with_antiparallel_z_axes_inverts_z_coord():
    q = quaternion_from_two_vectors((0.,0.,2.),(0.,0.,-2.))
    assert q.apply([1., 1., 1.])[2] == approx(-1., 1e-5)

def test_quaternion_from_two_vectors__with_almost_antiparallel_z_axes_mostly_inverts_yz_coords():
    q = quaternion_from_two_vectors((0.,0.,1.),(0.,0.00001,-0.99999))
    assert np.isclose(q.as_quat(), [-1., 0., 0., 0.], atol=1e-5).all()
    assert np.isclose(q.apply([1., 1., 1.]), [1., -1., -1.]).all()

def test_quaternion_from_two_vectors__with_parallel_axes_is_donothing_0001():
    q = quaternion_from_two_vectors((0., 0., 2.),(0., 0., 2.))
    assert np.isclose(q.as_quat(), [0., 0., 0., 1.]).all()
    assert (q.apply([1., 1., 1.]) == [1., 1., 1.]).all()

def test_quaternion_from_two_vectors__with_almost_parallel_axes_is_almost_donothing_0001():
    q = quaternion_from_two_vectors((0., 0., 1.),(0., 0.00001, 0.99999))
    assert np.isclose(q.as_quat(), [0., 0., 0., 1.], atol=1e-5).all()
    assert np.isclose(q.apply([1., 1., 1.]), [1., 1., 1.]).all()

def test_quaternion_from_two_vectors__with_perpendicular_xy_axes_rotates_90_degrees_around_z_axis():
    q = quaternion_from_two_vectors((1., 0., 0.),(0., 1., 0.))
    assert np.isclose(q.as_quat(), [0., 0., 2**0.5/2, 2**0.5/2.], atol=1e-5).all()
    assert np.isclose(q.apply([1., 1., 1.]), [-1., 1., 1.]).all()
    assert np.isclose(q.apply([-1., 1., 1.]), [-1., -1., 1.]).all()
    assert np.isclose(q.apply([-1., -1., 1.]), [1., -1., 1.]).all()
    assert np.isclose(q.apply([1., -1., 1.]), [1., 1., 1.]).all()

def test_quaternion_from_two_vectors_around_axis__with_parallel_vectors_is_donothing_0001():
    q = quaternion_from_two_vectors_around_axis((0., 0., 2.),(0., 0., 2.), axis=(1., 0., 0.))
    assert np.isclose(q.as_quat(), [0., 0., 0., 1.]).all()
    assert (q.apply([1., 1., 1.]) == [1., 1., 1.]).all()

def test_quaternion_from_two_vectors_around_axis__with_antiparallel_about_x_reverses_yz():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        q = quaternion_from_two_vectors_around_axis((-5, -1, 0.),(5, 1, 0.), axis=(1., 0., 0.))

    assert np.isclose(q.apply([1., 1., 1.]), [1., -1., -1.]).all()
    assert np.isclose(q.apply([1., -1., 1.]), [1., 1., -1.]).all()
    assert np.isclose(q.apply([1., 1., -1.]), [1., -1., 1.]).all()

def test_quaternion_from_two_vectors_around_axis__perp():
    q = quaternion_from_two_vectors_around_axis((0., 0., 1.), (0., 1., 0.), (1., 0., 0.))
    assert np.isclose(q.apply([0., 0., 1.]), [0., 1., 0.]).all()

    q = quaternion_from_two_vectors_around_axis((1., 0., 1.), (1., 1., 0.), (1., 0., 0.))
    assert np.isclose(q.apply([1., 0., 1.]), [1., 1., 0.]).all()



"""
Run and visualize with:

python -m cProfile -o hkustperf.pstats  -s tottime hkustperf.py
snakeviz hkustperf.pstats

"""


import ase
import ase.io

import numpy as np

from mofun import find_pattern_in_structure, replace_pattern_in_structure, Atoms


hkust1_cif = Atoms.load_cif("../tests/hkust-1/hkust-1-with-bonds.cif")
benzene = Atoms.from_ase_atoms(ase.io.read("../tests/molecules/benzene.xyz"))
hkust1_repped = hkust1_cif.replicate(repldims=(3,3,3))
match_indices = find_pattern_in_structure(hkust1_repped, benzene)



import re

def parse_gas_loading(output_file, units='molkg'):
    atom_blocks = []

    with open(output_file) as origin:
        lines = origin.read().split('\n')
        absvloading = None
        absvloading_error = None
        for i, line in enumerate(lines):
            if units=='molkg' and "Average loading absolute [mol/kg framework]" in line:
                absvloading = float(line.split()[5])
                absvloading_error = float(line.split()[7])
            elif units=='vv' and "Average loading absolute [cm^3 (STP)/cm^3 framework]" in line:
                absvloading = float(line.split()[6])
                absvloading_error = float(line.split()[8])
            elif units=='cc_g' and "Average loading absolute [cm^3 (STP)/gr framework]" in line:
                absvloading = float(line.split()[6])
                absvloading_error = float(line.split()[8])
            elif units=='molkg' and "Average loading excess [mol/kg framework]" in line:
                excessvloading = float(line.split()[5])
                excessvloading_error = float(line.split()[7])
            elif units=='vv' and "Average loading excess [cm^3 (STP)/cm^3 framework]" in line:
                excessvloading = float(line.split()[6])
                excessvloading_error = float(line.split()[8])
            elif units=='cc_g' and "Average loading excess [cm^3 (STP)/gr framework]" in line:
                excessvloading = float(line.split()[6])
                excessvloading_error = float(line.split()[8])
            elif "Number of molecules:" in line:
                atom_blocks = [float(lines[offset + i + 5].split()[2]) for offset in range(5)]
            elif "Conversion factor molecules/unit cell -> cm^3 STP/cm^3:" in line:
                molecules_uc_to_vv = float(line.split()[7])
            elif "Conversion factor mol/kg -> cm^3 STP/cm^3:" in line:
                molkg_to_vv = float(line.split()[6])

    return absvloading, absvloading_error, excessvloading, excessvloading_error, *atom_blocks, molecules_uc_to_vv, molkg_to_vv


def parse_henrys(output_file):
    fl = r"[-+]?\d+(?:\.\d*)?(?:[eE][-+]?\d+)?"

    # captured groups are [gas, Henry's coefficient, Henry's coefficient error]
    henrys_re = re.compile(r"\s+\[([-\w]+)\] Average Henry coefficient:  ({fl}) \+\/\- ({fl}) \[mol/kg/Pa\]".format(fl=fl))

    gas_henrys_error = []
    with open(output_file) as f:
        for line in f:
            m=re.match(henrys_re, line)
            if m:
                matches = list(m.groups())
                matches[1] = float(matches[1])
                matches[2] = float(matches[2])
                gas_henrys_error += [matches]
            elif "Conversion factor molecules/unit cell -> cm^3 STP/cm^3:" in line:
                molecules_uc_to_vv = float(line.split()[7])
            elif "Conversion factor mol/kg -> cm^3 STP/cm^3:" in line:
                molkg_to_vv = float(line.split()[6])
    return gas_henrys_error, molecules_uc_to_vv, molkg_to_vv



from collections import Counter
import itertools
from math import sqrt, log, cos, sin, pi

import networkx as nx
import numpy as np

from mofun.atomic_masses import ATOMIC_MASSES
from mofun.uff4mof import UFF4MOF, uff_key_starts_with, MAIN_GROUP_ELEMENTS
from mofun.helpers import typekey

def default_uff_rules():
    return {
        "H": [
            ("H_b", dict(n=2)),
            ("H_", {})
        ],
        "O": [
            ("O_1", dict(n=1))
        ],
        "C": [
            ("C_R", dict(aromatic=True))
        ],
        "N": [
            ("N_R", dict(aromatic=True))
        ]
    }

def add_aromatic_flag(g):
    cycles = nx.cycle_basis(g)
    for cycle in cycles:
        if 5 <= len(cycle) <= 7:
            for n in cycle:
                g.nodes[n]['aromatic'] = True

def retype_atoms_from_uff_types(atoms, new_types):
    """Retypes an atoms object with new UFF types.

    Takes a new list of types (new_types) of length equal to the number of atoms in the system and
    creates a new set of unique atom_types, where atom_type_labels are the UFF atom types, and atom_type_elements
    and atom_type_masses are the appropriate element and mass for the UFF atom type. New atom_types are
    assigned to the atoms object, based on the new atom type indices.

    Args:
        atoms (Atoms): atoms object to retype
        new_types (List[Str]): list of UFF atom types, one per atom in the system.
    """

    # sort by string ordering, so types like 'C_1', 'C_2', 'C_3', 'C_R' will show up in order
    unique_types = list(set(new_types))
    unique_types.sort()

    # sort by periodic element # order
    ptable_order = lambda x: list(ATOMIC_MASSES.keys()).index(x[0:2].replace('_', ''))
    unique_types.sort(key=ptable_order)

    atoms.atom_type_labels = unique_types
    atoms.atom_type_elements = [s[0:2].replace('_', '') for s in unique_types]
    atoms.atom_type_masses = [ATOMIC_MASSES[s] for s in atoms.atom_type_elements]

    atoms.atom_types = [unique_types.index(s) for s in new_types]

def calc_uff_atom_types(bonds, elements, override_rules=None):
    """
    """

    g = nx.Graph()
    g.add_edges_from(bonds)

    if override_rules is None:
        override_rules = default_uff_rules()

    uff_keys = UFF4MOF.keys()

    atom_types = []
    add_aromatic_flag(g)
    for n in sorted(g.nodes):
        # handle override rules
        el = elements[n]
        found_type = False

        if el == "C":
            h = g.degree(n) - 1
        elif el == "N":
            h = g.degree(n)
        elif el == "O":
            if g.degree(n) == 1:
                h = 2
            else:
                h = 3
        else:
            h = g.degree(n) - 1

        if el in override_rules:
            for ufftype, reqs in override_rules[el]:
                if "n" in reqs and g.degree(n) != reqs['n']:
                    continue
                if "h" in reqs and h != reqs['h']:
                    continue
                if "aromatic" in reqs and "aromatic" not in g.nodes[n]:
                    continue
                if "neighbors" in reqs and \
                        sorted([elements[i] for i in g.neighbors(n)]) != sorted(reqs['neighbors']):
                    continue
                found_type = True
                atom_types.append(ufftype)
                break
            if found_type:
                continue

        # handle default cases
        # 1: try element + hybridization
        uff_key = "%s%1d" % (el.ljust(2, "_"), h)
        possible_uff_keys = uff_key_starts_with(uff_key)
        if len(possible_uff_keys) == 1:
            atom_types.append(possible_uff_keys[0])
            continue
        elif len(possible_uff_keys) >= 2:
            if uff_key in UFF4MOF:
                atom_types.append(uff_key)
                print("Warning: multiple possible UFF keys: %s. Choosing the simplest one: %s" % (possible_uff_keys, uff_key))
                continue
            else:
                raise Exception("Error: too many possible UFF keys that starts with %s with no way of discerning which one: %s" % (uff_key, possible_uff_keys))

        # 2: try element w/o hybridization
        # Note that UFF4MOF adds some fancy types, i.e. Li3f2, and if the hybridization doesn't match
        # above, then this will still default to the original UFF Li term.
        if (el := uff_key[0:2]) in UFF4MOF:
            atom_types.append(el)
            continue

        raise Exception("no appropriate UFF key that starts with %s" % uff_key)

    return atom_types

def pair_coeffs(a1):
    lj_sigma = UFF4MOF[a1][2] * (2**(-1./6.))
    lj_epsilon = UFF4MOF[a1][3]
    return [lj_epsilon, lj_sigma]

def guess_bond_order(a1, a2, rules=[]):
    """
    Args:
        a1 (str): UFF atom type of atom 1 in bond 1-2, e.g. "C_R"
        a2 (str): UFF atom type of atom 2 in bond 1-2, e.g. "C_R"
        rules (optional): a list of tuple pairs where each pair contains a set of atom_types
            representing a bond, and a bo. E.g.: [({N_1}, 2), ({N_1, N_2}, 2)] where that means a
            bond between two N_1 atoms would have a bond order of 2, and a bond between an N_1 and a
            N_2 would also have a bond_order of 2.

    This method is 'hacky' at best and could be replaced by something more sophisticated.
    This is roughly the same as what is used in Pete Boyd's 'lammps-interface'.
    """
    bond_atom_types = {a1, a2}

    for rule_atom_types, bo in rules:
        if bond_atom_types == rule_atom_types:
            return bo

    if len({'H_', 'F_', 'Cl', 'Br', 'I_', 'C_3', 'N_3', 'O_3'} & bond_atom_types) > 0:
        return 1
    elif len(bond_atom_types) == 1 and bond_atom_types <= {'C_2', 'N_2', 'O_2'}:
        return 2
    elif len(bond_atom_types) == 1 and bond_atom_types <= {'C_R', 'N_R', 'O_R'}:
        return 1.5
    else:
        print('Warning: %s %s Bond order not explicitly assigned. Using default value of 1.' % (a1, a2))
        return 1

def bond_params(a1, a2, bond_order=None, bond_order_rules=[]):
    """calculate the UFF bond parameters in a form suitable for calculations in LAMMPS.

    Args:
        a1 (str): UFF atom type of atom 1 in bond 1-2, e.g. "C_R"
        a2 (str): UFF atom type of atom 2 in bond 1-2, e.g. "C_R"
        bond_order (float): bond order for the bond

    Returns:
        (float, float): tuple of bond force constant and bond length.
    """

    if bond_order is None:
        bond_order = guess_bond_order(a1, a2, bond_order_rules)

    # 1. Calculate rij and Kij
    ri, zi, chii = [UFF4MOF[a1][k] for k in (0, 5, 8)]
    rj, zj, chij = [UFF4MOF[a2][k] for k in (0, 5, 8)]

    rBO = -0.1332 * (ri + rj) * log(bond_order)
    rEN = (ri * rj * (chii**0.5 - chij**0.5)**2) / (chii * ri + chij * rj)

    rij = ri + rj + rBO - rEN
    kij = 664.12 * zi * zj / (rij**3)

    return (kij / 2, rij)

def angle_params(a1, a2, a3, bond_orders=[None, None], bond_order_rules=[]):
    """calculate the UFF angle parameters in a form suitable for calculations in LAMMPS.

    Args:
        a1 (str): UFF atom type of atom 1 in angle 1-2-3, e.g. "C_R"
        a2 (str): UFF atom type of atom 2 in angle 1-2-3, e.g. "C_R"
        a3 (str): UFF atom type of atom 3 in angle 1-2-3, e.g. "C_R"
        a2_coord_num (int): coordination # of a2 atom.
        bond_orders ([float, float]): bond orders for the 1-2 and 2-3 bonds.

    Returns:
        (str, float, int, int): For linear cases, returns a tuple of ('cosine/periodic', C, b, n)
            where C, b, n are defined by https://lammps.sandia.gov/doc/angle_cosine_periodic.html.
        (str, float, float, float): For the nonlinear general case, returns a tuple of
            ('fourier', K, c0, c1, c2) where the args are defined by
            https://lammps.sandia.gov/doc/angle_fourier.html
    """

    a2_coord_is_4 = (a2[2] == "3") if len(a2) > 2 else False

    theta0deg = UFF4MOF[a2][1]
    theta0rad = theta0deg * 2 * pi / 360

    # Determine force constant
    rij = bond_params(a1, a2, bond_order=bond_orders[0], bond_order_rules=bond_order_rules)[1]
    rjk = bond_params(a2, a3, bond_order=bond_orders[1], bond_order_rules=bond_order_rules)[1]
    rik = sqrt(rij**2 + rjk**2 - 2 * rij * rjk * cos(theta0rad))

    zi = UFF4MOF[a1][5]
    zk = UFF4MOF[a3][5]
    kijk = 664.12 * (zi * zk / rik**5) * (3 * rij * rjk * (1 - cos(theta0rad)**2) -
            (rik**2 * cos(theta0rad)))

    if theta0deg in [180., 120., 90.]:
        # Linear cases use a simplified fourier expansion. This is written in the Rappe paper as:
        # E = Kijk/n^2[1 - cos(n*theta)]
        # The corresponding term in LAMMPS is a cosine/periodic term of the form:
        # E = C[1 - B(-1)^n cos(n*theta)]

        if theta0deg == 180.:
            n = 1
            # use 1 + cos x
            b = 1
        elif theta0deg == 120.:
            n = 3
            b = -1
        elif theta0deg == 90. and a2_coord_is_4:
            n = 2
            # use 1 + cos x
            b = -1
        elif theta0deg == 90.:
            n = 4
            b = 1

        return ('cosine/periodic', kijk, b, n)
    else:
        # General nonlinear cases use a three-term fourier expansion of the form:
        # E = Kijk * [C0 + C1 cos(theta) + c2 cos(2*theta)]
        # The corresponding term in LAMMPS is a fourier term.

        c2 = 1 / (4 * sin(theta0rad)**2)
        c1 = -4 * c2 * cos(theta0rad)
        c0 = c2 * (2 * cos(theta0rad)**2 + 1)
        return ('fourier', kijk, c0, c1, c2)

def dihedral_params(a1, a2, a3, a4, num_dihedrals_about_bond=1, bond_order=None, bond_order_rules=[]):
    """Use a small cosine Fourier expansion

    E_phi = 1/2*V_phi * [1 - cos(n*phi0)*cos(n*phi)]

    this is available in Lammps in the form of a harmonic potential
    E = K * [1 + d*cos(n*phi)]

    K = V_phi / 2
    d = -cos(n*phi0)
    n = n

    NB: the d term must be negated to recover the UFF potential.
    """

    # get the elements and hybridizations from each UFF atom type
    ut = [a1, a2, a3, a4]
    el = [s[0:2].strip('_') for s in ut]
    h = [s[2] if len(s) > 2 else 0 for s in ut]

    if bond_order is None:
        bond_order = guess_bond_order(a2, a3, bond_order_rules)

    oxygen_group = {'O', 'S', 'Se', 'Te', 'Po'}

    print("dihedral: %s-%s-%s-%s M=%d" % (*tuple(ut), num_dihedrals_about_bond))
    if {h[1], h[2]} <= {'3'}:
        print("sp3-sp3")
        # center atoms are sp3, use eq 16 from Rappe
        # n=3, phi0 = 180 or 60
        # cos(n*phi0) = cos(3*180), cos(3*60) => cos(540), cos(180) = -1
        # for oxy exception: cos(2*90) = cos(180) = -1
        # hence, d = 1

        n = 3
        v1 = UFF4MOF[ut[1]][6]
        v2 = UFF4MOF[ut[2]][6]

        # exception for when both atoms are from group 6
        if {el[1], el[2]} <= oxygen_group:
            print("exception: both oxys")
            n = 2
            v1 = 2. if el[1] == "O" else 6.8
            v2 = 2. if el[2] == "O" else 6.8

        v = sqrt(v1*v2) / num_dihedrals_about_bond
        return ("harmonic", v/2, 1, n)

    elif {h[1], h[2]} <= {'2', 'R'}:
        print("sp2-sp2")
        # center atoms are sp2 (or aromatic), use eq 17 from Rappe,
        # phi0=180, n=2, cos(2*180) = 1, hence d = -1
        v = 5.0 * sqrt(UFF4MOF[a2][7] * UFF4MOF[a3][7]) * (1. + 4.18 * log(bond_order)) / num_dihedrals_about_bond
        n = 2
        return ("harmonic", v/2, -1, n)

    elif {h[1], h[2]} <= {'2', 'R', '3'}:
        print("sp2-sp3")
        # mixed sp2 / sp3 / aromatic case
        if {h[0], h[1]} <= {'2'} or {h[2], h[3]} <= {'2'}:
            print("exception: sp2 to another sp2")
            # exception for when the sp2 is bonded to another sp2,
            # n = 3, phi0 = 180, cos(3*180) = -1, d=1

            n = 3
            v = 2. / num_dihedrals_about_bond
            return ("harmonic", v/2, 1, n)

        if (h[1] == '3' and el[1] in oxygen_group and el[2] not in oxygen_group) or \
             (h[2] == '3' and el[2] in oxygen_group and el[1] not in oxygen_group):
            print("exception: sp3 oxy, sp2/resonant other")
            # exception for sp3 from oxygen column and sp2 or resonant from another column
            # phi0=90, n=2, cos(2*90) = -1, d = 1
            # use eq 17 from rappe
            v = 5.0 * sqrt(UFF4MOF[ut[1]][7] * UFF4MOF[ut[2]][7]) * (1. + 4.18 * log(bond_order)) / num_dihedrals_about_bond
            n = 2
            return ("harmonic", v/2, 1, n)

        # default mixed sp2 / sp3 case
        # phi0 = 0, cos 6*0 = 1, d = -1
        n = 6
        v = 1. / num_dihedrals_about_bond
        return ("harmonic", v/2, -1, n)
    elif '1' in {h[1], h[2]}:
        # no dihedrals for "sp-hybridized centers X_1"
        return None
    elif not {el[1], el[2]} <= set(MAIN_GROUP_ELEMENTS):
        # no dihedrals for non main group elements
        return None
    else:
        raise Exception("we don't know how to handle this dihedral: %s-%s-%s-%s" % tuple(ut))

def calc_angles(bonds):
    """ Returns all possible angle tuples from a list of all system bond tuples."""
    g = nx.Graph()
    g.add_edges_from(bonds)

    angles = []
    for n in g.nodes:
        angles += [(a, n, b) for (a,b) in itertools.combinations(g.neighbors(n), 2)]
    return np.array(angles)

def calc_dihedrals(bonds):
    """ Returns all possible dihedral tuples from a list of all system bond tuples."""
    g = nx.Graph()
    g.add_edges_from(bonds)
    # g.add_edges_from([tuple(x) for x in bonds])

    dihedrals = []
    for a, b in g.edges:
        a_neighbors = list(g.adj[a])
        a_neighbors.remove(b)
        b_neighbors = list(g.adj[b])
        b_neighbors.remove(a)

        dihedrals += [(a1, a, b, b1) for a1 in a_neighbors for b1 in b_neighbors]
    return np.array(dihedrals)

def delete_if_all_in_set(arr, s):
    deletion_list = []
    for i, tup in enumerate(arr):
        if len(set(tup) - s) == 0:
            deletion_list.append(i)
    return np.delete(arr, deletion_list, axis=0)

def assign_pair_coeffs(atoms, assign_atom_type_labels_from_elements=False):
    if assign_atom_type_labels_from_elements:
        atoms.atom_type_labels = [uff_key_starts_with(el.ljust(2, "_"))[0] for el in atoms.atom_type_elements]

    atoms.pair_coeffs = ['%10.6f %10.6f # %s' % (*pair_coeffs(a1), a1) for a1 in atoms.atom_type_labels]

def assign_bond_types(atoms, uff_atom_types, bond_order_rules=[], exclude=[]):
    if len(exclude) >= 2:
        atoms.bonds = delete_if_all_in_set(atoms.bonds, exclude)

    # get an ordered list of bond types for every bond, e.g. [(1,2), (4,5)]
    bond_types = [typekey([uff_atom_types[a] for a in atup]) for atup in atoms.bonds]
    # reduce the list so every type is unique
    unique_bond_types = list(dict.fromkeys(bond_types).keys())
    # bond_types are the index of the type in the unique_bond_types list
    atoms.bond_types = [unique_bond_types.index(bt) for bt in bond_types]
    # calculate bond_params from bond types and bond_order_rules and assign
    params = [(*bond_params(a1, a2, bond_order_rules=bond_order_rules), a1, a2) for (a1, a2) in unique_bond_types]
    atoms.bond_type_coeffs = ['%10.6f %10.6f # %s %s' % p for p in params]

def angle2lammpsdat(params):
    if params[0] == "fourier":
        return '%s %10.6f %10.6f %10.6f %10.6f # %s' % params
    elif params[0] == "cosine/periodic":
        return '%s %10.6f %d %d # %s' % params
    else:
        raise Exception("Unhandled angle style '%s'" % params[0])

def assign_angle_types(atoms, uff_atom_types, bond_order_rules=[], exclude=[]):
    if len(exclude) >= 3:
        atoms.angles = delete_if_all_in_set(atoms.angles, exclude)

    angle_types = [typekey([uff_atom_types[a] for a in atup]) for atup in atoms.angles]
    unique_angle_types = list(dict.fromkeys(angle_types).keys())
    atoms.angle_types = [unique_angle_types.index(a) for a in angle_types]
    params = [(*angle_params(*a_ids, bond_order_rules=bond_order_rules), "%s %s %s" % a_ids) for a_ids in unique_angle_types]
    atoms.angle_type_coeffs = [angle2lammpsdat(a) for a in params]

def assign_dihedral_types(atoms, uff_atom_types, bond_order_rules=[], exclude=[]):
    """

    Args:
        exclude (List): removes dihedrals from list if the dihedral's atom
            indices are ALL contained within exclude.
    """
    num_dihedrals_per_bond = Counter([typekey([a2, a3]) for _, a2, a3, _ in atoms.dihedrals])
    if len(exclude) >= 4:
        atoms.dihedrals = delete_if_all_in_set(atoms.dihedrals, exclude)

    # dihedral type is a tuple of four atom indices, and the number of dihedrals about the center bond
    dihedral_types = [(*typekey([uff_atom_types[a] for a in atup]), num_dihedrals_per_bond[typekey([atup[1], atup[2]])]) for atup in atoms.dihedrals]
    unique_dihedral_types = list(dict.fromkeys(dihedral_types).keys())
    # params = [dihedral_params(*a_ids, bond_order_rules=bond_order_rules) for a_ids in unique_dihedral_types]
    params = [(dihedral_params(*a_ids, bond_order_rules=bond_order_rules), a_ids) for a_ids in unique_dihedral_types]

    # delete any dihedrals when the params come back None (i.e. for *_1)
    for i in reversed(range(len(params))):
        if params[i][0] is None:
            # delete from atoms.dihedrals and dihedral_types
            d_to_del = unique_dihedral_types[i]
            atoms.dihedrals = [d for j, d in enumerate(atoms.dihedrals) if dihedral_types[j] != d_to_del]
            dihedral_types = [d for d in dihedral_types if d != d_to_del]
            # delete from params and types
            del(unique_dihedral_types[i])
            del(params[i])

    # assign dihedral types
    atoms.dihedral_types = [unique_dihedral_types.index(a) for a in dihedral_types]
    atoms.dihedral_type_coeffs = ['%s %10.6f %d %d # %s %s %s %s M=%d' % (*p1, *p2) for p1, p2 in params]



from contextlib import contextmanager
import functools
import math
import random
import warnings

import numpy as np
from scipy.linalg import norm
from scipy.spatial.transform import Rotation as R

from mofun.atomic_masses import ATOMIC_MASSES

def suppress_warnings(func):
    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return func(*args, **kwargs)
    return wrapper_decorator

def atoms_of_type(types, element):
    """ returns all atom indices in types that match the symbol element """
    return [i for i, t in enumerate(types) if t == element]

def atoms_by_type_dict(atom_types):
    atoms_by_type = {k:[] for k in set(atom_types)}
    for i, k in enumerate(atom_types):
        atoms_by_type[k].append(i)
    return atoms_by_type

def group_duplicates(match_indices, key=lambda m: tuple(sorted(m))):
    keyed_tuples = {}
    for m in match_indices:
        mkey = key(m)
        if mkey not in keyed_tuples:
            keyed_tuples[mkey] = [m]
        else:
            keyed_tuples[mkey].append(m)
    return keyed_tuples

def remove_duplicates(match_indices, key=lambda m: tuple(sorted(m)), pick_random=False):
    keyed_tuples = {}
    for m in match_indices:
        mkey = key(m)
        if mkey not in keyed_tuples:
            keyed_tuples[mkey] = [m]
        else:
            keyed_tuples[mkey].append(m)
    if pick_random:
        return [random.choice(matches) for _, matches in keyed_tuples.items()]
    else: # pick first
        return [matches[0] for _, matches in keyed_tuples.items()]

def position_index_farthest_from_axis(axis, atoms):
    q = quaternion_from_two_vectors(axis, [1., 0., 0.])
    ratoms = q.apply(atoms.positions)
    ss = (ratoms[:,1:3] ** 2).sum(axis=1)
    return np.nonzero(ss==ss.max())[0][0]

def quaternion_from_two_vectors(p1, p2):
    """ returns the quaternion necessary to rotate p1 to p2"""
    v1 = np.array(p1) / norm(p1)
    v2 = np.array(p2) / norm(p2)

    angle = np.arccos(max(-1.0, min(np.dot(v1, v2), 1)))
    axis = np.cross(v1, v2)

    if np.isclose(axis, [0., 0., 0.], 1e-3).all() and angle != 0.0:
        # the antiparallel case requires we arbitrarily find a orthogonal rotation axis, since the
        # cross product of a two parallel / antiparallel vectors is 0.
        axis = np.cross(v1, np.random.random(3))

    if norm(axis) > 1e-15:
        axis /= norm(axis)
    return R.from_quat([*(axis*np.sin(angle / 2)), np.cos(angle/2)])

def quaternion_from_two_vectors_around_axis(p1, p2, axis):
    """ returns the quaternion necessary to rotate p1 to p2"""
    axis = np.array(axis)

    # convert p1 / p2 to be vectors orthogonal to axis
    p1 = p1 - (np.dot(p1, axis) / np.dot(axis, axis)) * axis
    p2 = p2 - (np.dot(p2, axis) / np.dot(axis, axis)) * axis

    v1 = np.array(p1) / norm(p1)
    v2 = np.array(p2) / norm(p2)

    angle = np.arccos(max(-1.0, min(np.dot(v1, v2), 1)))

    if norm(axis) > 1e-15:
        axis /= norm(axis)

    if angle not in [0., math.pi] and np.isclose(axis, np.cross(v1, v2) / norm(np.cross(v1, v2)), 1e-3).all():
        angle *= -1
    return R.from_quat([*(axis*np.sin(-angle / 2)), np.cos(-angle/2)])

def guess_elements_from_masses(masses, max_delta=1e-2):
    def find_element(elmass):
        for sym, mass in ATOMIC_MASSES.items():
            if elmass - mass < max_delta:
                return sym
        raise Exception("no element matching mass %8.5f in elements list. Please add one?")

    return [find_element(m) for m in masses]

@contextmanager
def use_or_open(fh, path, mode='r'):
    if fh is None:
        with open(path, mode) as f:
            yield f
    else:
        yield fh

def typekey(tup):
    rev = list(tup)
    rev.reverse()
    if tuple(rev) <= tuple(tup):
        return tuple(rev)
    return tuple(tup)

class PositionsNotEquivalent(Exception):
    pass

def assert_structure_positions_are_unchanged(orig_structure, final_structure, max_delta=1e-5, verbose=True):
    return assert_positions_are_unchanged(orig_structure.positions, final_structure.positions, max_delta, verbose)

def assert_positions_are_unchanged(p, new_p, max_delta=1e-5, verbose=True, raise_exception=False):
    if raise_exception:
        if not positions_are_unchanged(p, new_p, max_delta, verbose):
            raise PositionsNotEquivalent()
    else:
         assert positions_are_unchanged(p, new_p, max_delta, verbose)

def positions_are_unchanged(p, new_p, max_delta=1e-5, verbose=True):
    p_ordered = p[np.lexsort((p[:,0], p[:,1], p[:,2]))]
    new_p_ordered = new_p[np.lexsort((new_p[:,0], new_p[:,1], new_p[:,2]))]

    if verbose:
        print("** positions_are_unchanged? **")
        print("p = \n", p)
        print("new_p = \n", new_p)
        print("p     (sorted) = \n", p_ordered)
        print("new_p (sorted) = \n", new_p_ordered)
    p_matched = []
    p_corresponding = []
    distances = np.full(len(p), max(9.99, 9.99 * max_delta))
    for i, p1 in enumerate(p_ordered):
        found_match = False
        for j, p2 in enumerate(new_p_ordered):
            # print(p2, p1, norm(np.array(p2) - p1))
            if p2[2] - p1[2] > 1:
                break
            elif (np21 := norm(np.array(p2) - p1)) < max_delta:
                found_match = True
                p_corresponding.append(new_p_ordered[j, :])
                new_p_ordered = np.delete(new_p_ordered, j, axis=0)
                p_matched.append(i)
                distances[i] = np21
                break
        if not found_match:
            p_corresponding.append([])

    p_unmatched = np.delete(p_ordered, p_matched, 0)
    distances = np.array(distances)
    if verbose:
        for i, p1 in enumerate(p_ordered):
            annotation = ""
            if distances[i] > max_delta:
                annotation = " * "
            print(i, p1, p_corresponding[i], distances[i], annotation)
        print("UNMATCHED coords in old positions: ")
        for p1 in p_unmatched:
            print(p1)
        print("UNMATCHED coords in new positions: ")
        for p1 in new_p_ordered:
            print(p1)
        print("--")
    return (distances < max_delta).all()



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(name='mofun',
      version='0.1.1',
      description='Find and replace functional groups in any given periodic structure.',
      author='Paul Boone',
      author_email='narcissus.pong@gmail.com',
      url='https://github.com/wilmerlab/mofun.git',
      packages=find_packages(include=['mofun*']),
      install_requires=[
          'ase',
          'numpy',
          'scipy',
          'PyCifRW',
          'networkx',
          'click',
      ],
      extras_require={
        'docs': ['mkdocs-material', 'mkdocstrings'],
      },
      entry_points={
        'console_scripts': [
            'mofun = mofun.cli.mofun_cli:mofun_cli',
          ]
      }

)
